# Copyright (C) 2013-2016 Kristoffer Gronlund <kgronlund@suse.com>
# See COPYING for license information.

"""
Helpers for handling log timestamps.
"""

import re
import time
import datetime
from . import utils
from . import log


logger = log.setup_logger(__name__)


YEAR = None


def set_year(ts=None):
    '''
    ts: optional time in seconds
    '''
    global YEAR
    year = time.strftime("%Y", time.localtime(ts))
    if YEAR is not None:
        t = (" (ts: %s)" % (ts)) if ts is not None else ""
        logger.debug("history: setting year to %s%s", year, t)
    YEAR = year


def human_date(dt=None):
    '''
    Convert datetime argument into a presentational string.

    dt: Datetime (default: now)
    '''
    if dt is None:
        dt = utils.make_datetime_naive(datetime.datetime.now())
    # here, dt is in UTC. Convert to localtime:
    localdt = datetime.datetime.fromtimestamp(utils.datetime_to_timestamp(dt))
    # drop microseconds
    return re.sub("[.].*", "", "%s %s" % (localdt.date(), localdt.time()))


def make_time(t):
    '''
    t: time in seconds / datetime / other
    returns: time in floating point
    '''
    if t is None:
        return None
    elif isinstance(t, datetime.datetime):
        return utils.datetime_to_timestamp(t)
    return t


# fmt1: group 11 is node
# fmt2: group 2 is node
# fmt3: group 2 is node
# fmt4: node not available?
_syslog2node_formats = (re.compile(r'^(\d{4})-(\d{2})-(\d{2})T(\d{2}):(\d{2}):(\d{2})(?:.(\d+))?([+-])(\d{2}):?(\d{2})\s+(?:\[\d+\])?\s*([\S]+)'),
                        re.compile(r'^(\d{4}-\d{2}-\d{2}T\S+)\s+(?:\[\d+\])?\s*([\S]+)'),
                        re.compile(r'^([a-zA-Z]{2,4}\s+\d{1,2}\s+\d{2}:\d{2}:\d{2})\s+(?:\[\d+\])?\s*([\S]+)'),
                        re.compile(r'^(\d{4}\/\d{2}\/\d{2}_\d{2}:\d{2}:\d{2})'),
                        re.compile(r'^([A-Z][a-z]+ \d{1,2} \d{2}:\d{2}:\d{2}\.\d+) ([\S]+)'))

_syslog_ts_prev = None


def syslog_ts(s):
    """
    Finds the timestamp in the given line
    Returns as floating point, seconds
    """
    global _syslog_ts_prev
    fmt1, fmt2, fmt3, fmt4, fm5 = _syslog2node_formats

    # RFC3339
    m = fmt1.match(s)
    if m:
        year, month, day, hour, minute, second, ms, tzsgn, tzh, tzm, _ = m.groups()
        ts = time.mktime((int(year), int(month), int(day), int(hour), int(minute), int(second), 0, 0, -1))
        if tzsgn == '+':
            ts += (3600.0 * float(tzh) + 60.0 * float(tzm))
        else:
            ts -= (3600.0 * float(tzh) + 60.0 * float(tzm))
        if ms:
            ts += float("0.%s" % ms)
        _syslog_ts_prev = ts
        return _syslog_ts_prev

    m = fmt2.match(s)
    if m:
        _syslog_ts_prev = utils.parse_to_timestamp(m.group(1))
        return _syslog_ts_prev

    m = fmt3.match(s)
    if m:
        if YEAR is None:
            set_year()
        tstr = YEAR + ' ' + m.group(1)

        dt = datetime.datetime.strptime(tstr, '%Y %b %d %H:%M:%S')
        from dateutil import tz
        ts = utils.total_seconds(dt - tz.tzlocal().utcoffset(dt) - datetime.datetime(1970, 1, 1))
        _syslog_ts_prev = ts
        return _syslog_ts_prev

    m = fmt4.match(s)
    if m:
        tstr = m.group(1).replace('_', ' ')
        _syslog_ts_prev = utils.parse_to_timestamp(tstr)
        return _syslog_ts_prev

    m = fm5.match(s)
    if m:
        _syslog_ts_prev = utils.parse_to_timestamp(m.group(1))
        return _syslog_ts_prev

    logger.debug("malformed line: %s", s)
    return _syslog_ts_prev


_syslog_node_prev = None


def syslog2node(s):
    '''
    Get the node from a syslog line.

    old format:
    Aug 14 11:07:04 <node> ...
    new format:
    Aug 14 11:07:04 [<PID>] <node> ...
    RFC5424:
    <TS> <node> ...
    RFC5424 (2):
    <TS> [<PID>] <node> ...
    '''
    global _syslog_node_prev

    fmt1, fmt2, fmt3, _, _ = _syslog2node_formats
    m = fmt1.match(s)
    if m:
        _syslog_node_prev = m.group(11)
        return _syslog_node_prev

    m = fmt2.match(s)
    if m:
        _syslog_node_prev = m.group(2)
        return _syslog_node_prev

    m = fmt3.match(s)
    if m:
        _syslog_node_prev = m.group(2)
        return _syslog_node_prev

    try:
        # strptime defaults year to 1900 (sigh)
        time.strptime(' '.join(s.split()[0:3]),
                      "%b %d %H:%M:%S")
        _syslog_node_prev = s.split()[3]
        return _syslog_node_prev
    except ValueError:  # try the rfc5424
        ls = s.split()
        if not ls:
            return _syslog_node_prev
        rfc5424 = s.split()[0]
        if 'T' in rfc5424:
            try:
                utils.parse_to_timestamp(rfc5424)
                _syslog_node_prev = s.split()[1]
                return _syslog_node_prev
            except Exception:
                return _syslog_node_prev
        else:
            return _syslog_node_prev


def syslog_ts_node(s):
    """
    Returns (timestamp, node) from a syslog log line
    """
    global _syslog_ts_prev
    global _syslog_node_prev
    fmt1, fmt2, fmt3, fmt4, fmt5 = _syslog2node_formats

    # RFC3339
    m = fmt1.match(s)
    if m:
        year, month, day, hour, minute, second, ms, tzsgn, tzh, tzm, node = m.groups()
        ts = time.mktime((int(year), int(month), int(day), int(hour), int(minute), int(second), 0, 0, -1))
        if tzsgn == '+':
            ts += (3600.0 * float(tzh) + 60.0 * float(tzm))
        else:
            ts -= (3600.0 * float(tzh) + 60.0 * float(tzm))
        _syslog_ts_prev = ts
        _syslog_node_prev = node
        return _syslog_ts_prev, node

    m = fmt2.match(s)
    if m:
        _syslog_ts_prev, _syslog_node_prev = utils.parse_to_timestamp(m.group(1)), m.group(2)
        return _syslog_ts_prev, _syslog_node_prev

    m = fmt3.match(s)
    if m:
        if YEAR is None:
            set_year()
        tstr = YEAR + ' ' + m.group(1)

        dt = datetime.datetime.strptime(tstr, '%Y %b %d %H:%M:%S')
        from dateutil import tz
        ts = utils.total_seconds(dt - tz.tzlocal().utcoffset(dt) - datetime.datetime(1970, 1, 1))
        _syslog_ts_prev, _syslog_node_prev = ts, m.group(2)
        return _syslog_ts_prev, _syslog_node_prev

    m = fmt4.match(s)
    if m:
        tstr = m.group(1).replace('_', ' ')
        _syslog_ts_prev = utils.parse_to_timestamp(tstr)
        return _syslog_ts_prev, _syslog_node_prev

    m = fmt5.match(s)
    if m:
        _syslog_ts_prev, _syslog_node_prev = utils.parse_to_timestamp(m.group(1)), m.group(2)
        return _syslog_ts_prev, _syslog_node_prev

    logger.debug("malformed line: %s", s)
    return _syslog_ts_prev, _syslog_node_prev



import os
import re
import socket
import functools
import typing
from enum import Enum

import crmsh.parallax
from . import sh
from . import utils
from . import parallax
from . import corosync
from . import xmlutil
from . import bootstrap
from . import lock
from . import log
from .service_manager import ServiceManager
from .sbd import SBDManager, SBDTimeout, SBDUtils


logger = log.setup_logger(__name__)
logger_utils = log.LoggerUtils(logger)


QDEVICE_ADD = "add"
QDEVICE_REMOVE = "remove"


class QdevicePolicy(Enum):
    QDEVICE_RELOAD = 0
    QDEVICE_RESTART = 1
    QDEVICE_RESTART_LATER = 2


def evaluate_qdevice_quorum_effect(mode, diskless_sbd=False, is_stage=False):
    """
    While adding/removing qdevice, get current expected votes and actual total votes,
    to calculate after adding/removing qdevice, whether cluster has quorum
    return different policy
    """
    quorum_votes_dict = utils.get_quorum_votes_dict()
    expected_votes = int(quorum_votes_dict["Expected"])
    actual_votes = int(quorum_votes_dict["Total"])
    if mode == QDEVICE_ADD:
        expected_votes += 1
    elif mode == QDEVICE_REMOVE:
        actual_votes -= 1

    if utils.calculate_quorate_status(expected_votes, actual_votes) and not diskless_sbd:
        # safe to use reload
        return QdevicePolicy.QDEVICE_RELOAD
    elif mode == QDEVICE_ADD and not is_stage:
        # Add qdevice from init process, safe to restart
        return QdevicePolicy.QDEVICE_RESTART
    elif xmlutil.CrmMonXmlParser().is_non_stonith_resource_running():
        # will lose quorum, with non-stonith resource running
        # no reload, no restart cluster service
        # just leave a warning
        return QdevicePolicy.QDEVICE_RESTART_LATER
    else:
        # will lose quorum, without resource running or just stonith resource running
        # safe to restart cluster service
        return QdevicePolicy.QDEVICE_RESTART


def qnetd_lock_for_same_cluster_name(func):
    """
    Decorator to claim lock on qnetd, to avoid the same cluster name added in qnetd
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        cluster_name = args[0].cluster_name
        lock_dir = "/run/.crmsh_qdevice_lock_for_{}".format(cluster_name)
        lock_inst = lock.RemoteLock(args[0].qnetd_addr, for_join=False, lock_dir=lock_dir, wait=False)
        try:
            with lock_inst.lock():
                func(*args, **kwargs)
        except lock.ClaimLockError:
            utils.fatal("Duplicated cluster name \"{}\"!".format(cluster_name))
        except lock.SSHError as err:
            utils.fatal(err)
    return wrapper


def qnetd_lock_for_multi_cluster(func):
    """
    Decorator to claim lock on qnetd, to avoid possible race condition
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        lock_inst = lock.RemoteLock(args[0].qnetd_addr, for_join=False, no_warn=True)
        try:
            with lock_inst.lock():
                func(*args, **kwargs)
        except (lock.SSHError, lock.ClaimLockError) as err:
            utils.fatal(err)
    return wrapper


class QDevice(object):
    """Class to manage qdevice configuration and services

    Call `certificate_process_on_init` to generate all of CA, server, and client certs.
    Call `certificate_process_on_join` to generate a single client cert for the local node.
    """

    qnetd_service = "corosync-qnetd.service"
    qnetd_cacert_filename = "qnetd-cacert.crt"
    qdevice_crq_filename = "qdevice-net-node.crq"
    qdevice_p12_filename = "qdevice-net-node.p12"
    qnetd_path = "/etc/corosync/qnetd"
    qdevice_path = "/etc/corosync/qdevice/net"
    qdevice_db_path = "/etc/corosync/qdevice/net/nssdb"

    def __init__(self, qnetd_addr, port=5403, algo="ffsplit", tie_breaker="lowest",
            tls="on", ssh_user=None, cluster_node=None, cmds=None, mode=None, cluster_name=None, is_stage=False):
        """
        Init function
        """
        self.qnetd_addr = qnetd_addr
        self.port = port
        self.algo = algo
        self.tie_breaker = tie_breaker
        self.tls = tls
        self.ssh_user = ssh_user
        self.cluster_node = cluster_node
        self.cmds = cmds
        self.mode = mode
        self.cluster_name = cluster_name
        self.qdevice_reload_policy = QdevicePolicy.QDEVICE_RESTART
        self.is_stage = is_stage
        self.using_diskless_sbd = False

    @property
    def qnetd_cacert_on_qnetd(self):
        """
        Return path of qnetd-cacert.crt on qnetd node
        """
        return "{}/nssdb/{}".format(self.qnetd_path, self.qnetd_cacert_filename)

    @property
    def qnetd_cacert_on_local(self):
        """
        Return path of qnetd-cacert.crt on local node
        """
        return "{}/{}/{}".format(self.qdevice_path, self.qnetd_addr, self.qnetd_cacert_filename)

    @property
    def qnetd_cacert_on_cluster(self):
        """
        Return path of qnetd-cacert.crt on cluster node
        """
        return "{}/{}/{}".format(self.qdevice_path, self.cluster_node, self.qnetd_cacert_filename)

    @property
    def qdevice_crq_on_qnetd(self):
        """
        Return path of qdevice-net-node.crq on qnetd node
        """
        return "{}/nssdb/{}.{}".format(self.qnetd_path, self.qdevice_crq_filename, self.cluster_name)
    
    @property
    def qdevice_crq_on_local(self):
        """
        Return path of qdevice-net-node.crq on local node
        """
        return "{}/nssdb/{}".format(self.qdevice_path, self.qdevice_crq_filename)

    @property
    def qnetd_cluster_crt_on_qnetd(self):
        """
        Return path of cluster-cluster_name.crt on qnetd node
        """
        return "{}/nssdb/cluster-{}.crt".format(self.qnetd_path, self.cluster_name)

    @property
    def qnetd_cluster_crt_on_local(self):
        """
        Return path of cluster-cluster_name.crt on local node
        """
        return "{}/{}/{}".format(self.qdevice_path, self.qnetd_addr, os.path.basename(self.qnetd_cluster_crt_on_qnetd))

    @property
    def qdevice_p12_on_local(self):
        """
        Return path of qdevice-net-node.p12 on local node
        """
        return "{}/nssdb/{}".format(self.qdevice_path, self.qdevice_p12_filename)

    @property
    def qdevice_p12_on_cluster(self):
        """
        Return path of qdevice-net-node.p12 on cluster node
        """
        return "{}/{}/{}".format(self.qdevice_path, self.cluster_node, self.qdevice_p12_filename)

    @staticmethod
    def check_qnetd_addr(qnetd_addr):
        qnetd_ip = None
        try:
            # socket.getaddrinfo works for both ipv4 and ipv6 address
            # The function returns a list of 5-tuples with the following structure:
            # (family, type, proto, canonname, sockaddr)
            # sockaddr is a tuple describing a socket address, whose format depends on the returned family
            # (a (address, port) 2-tuple for AF_INET, a (address, port, flow info, scope id) 4-tuple for AF_INET6)
            res = socket.getaddrinfo(qnetd_addr, None)
            qnetd_ip = res[0][-1][0]
        except socket.error:
            raise ValueError("host \"{}\" is unreachable".format(qnetd_addr))

        utils.node_reachable_check(qnetd_addr)

        if utils.InterfacesInfo.ip_in_local(qnetd_ip):
            raise ValueError("host for qnetd must be a remote one")

        if not utils.check_port_open(qnetd_ip, 22):
            raise ValueError("ssh service on \"{}\" not available".format(qnetd_addr))

    @staticmethod
    def check_qdevice_port(qdevice_port):
        if not utils.valid_port(qdevice_port):
            raise ValueError("invalid qdevice port range(1024 - 65535)")

    @staticmethod
    def check_qdevice_algo(qdevice_algo):
        if qdevice_algo not in ("ffsplit", "lms"):
            raise ValueError("invalid ALGORITHM choice: '{}' (choose from 'ffsplit', 'lms')".format(qdevice_algo))

    @staticmethod
    def check_qdevice_tie_breaker(qdevice_tie_breaker):
        if qdevice_tie_breaker not in ("lowest", "highest") and not utils.valid_nodeid(qdevice_tie_breaker):
            raise ValueError("invalid qdevice tie_breaker(lowest/highest/valid_node_id)")

    @staticmethod
    def check_qdevice_tls(qdevice_tls):
        if qdevice_tls not in ("on", "off", "required"):
            raise ValueError("invalid TLS choice: '{}' (choose from 'on', 'off', 'required')".format(qdevice_tls))

    @staticmethod
    def check_qdevice_heuristics_mode(mode):
        if not mode:
            return
        if mode not in ("on", "sync", "off"):
            raise ValueError("invalid MODE choice: '{}' (choose from 'on', 'sync', 'off')".format(mode))

    @staticmethod
    def check_qdevice_heuristics(cmds):
        if not cmds:
            return
        for cmd in cmds.strip(';').split(';'):
            if not cmd.startswith('/'):
                raise ValueError("commands for heuristics should be absolute path")
            if not os.path.exists(cmd.split()[0]):
                raise ValueError("command {} not exist".format(cmd.split()[0]))

    @staticmethod
    def check_package_installed(pkg, remote=None):
        if not utils.package_is_installed(pkg, remote_addr=remote):
            raise ValueError("Package \"{}\" not installed on {}".format(pkg, remote if remote else "this node"))

    def valid_qdevice_options(self):
        """
        Validate qdevice related options
        """
        self.check_package_installed("corosync-qdevice")
        self.check_qnetd_addr(self.qnetd_addr)
        self.check_qdevice_port(self.port)
        self.check_qdevice_algo(self.algo)
        self.check_qdevice_tie_breaker(self.tie_breaker)
        self.check_qdevice_tls(self.tls)
        self.check_qdevice_heuristics(self.cmds)
        self.check_qdevice_heuristics_mode(self.mode)

    def valid_qnetd(self):
        """
        Validate on qnetd node
        """
        exception_msg = ""
        suggest = ""
        shell = sh.cluster_shell()
        if not utils.package_is_installed("corosync-qnetd", remote_addr=self.qnetd_addr):
            exception_msg = "Package \"corosync-qnetd\" not installed on {}!".format(self.qnetd_addr)
            suggest = "install \"corosync-qnetd\" on {}".format(self.qnetd_addr)
        else:
            self.init_tls_certs_on_qnetd()
            self.start_qnetd()
            cmd = "corosync-qnetd-tool -l -c {}".format(self.cluster_name)
            if shell.get_stdout_or_raise_error(cmd, self.qnetd_addr):
                exception_msg = "This cluster's name \"{}\" already exists on qnetd server!".format(self.cluster_name)
                suggest = "consider to use the different cluster-name property"

        if exception_msg:
            if self.is_stage:
                exception_msg += "\nPlease {}.".format(suggest)
            else:
                exception_msg += "\nCluster service already successfully started on this node except qdevice service.\nIf you still want to use qdevice, {}.\nThen run command \"crm cluster init\" with \"qdevice\" stage, like:\n  crm cluster init qdevice qdevice_related_options\nThat command will setup qdevice separately.".format(suggest)
            raise ValueError(exception_msg)

    def enable_qnetd(self):
        ServiceManager().enable_service(self.qnetd_service, remote_addr=self.qnetd_addr)

    def disable_qnetd(self):
        ServiceManager().disable_service(self.qnetd_service, remote_addr=self.qnetd_addr)

    def start_qnetd(self):
        ServiceManager().start_service(self.qnetd_service, remote_addr=self.qnetd_addr)

    def stop_qnetd(self):
        ServiceManager().stop_service(self.qnetd_service, remote_addr=self.qnetd_addr)

    def set_cluster_name(self):
        if not self.cluster_name:
            self.cluster_name = corosync.get_value('totem.cluster_name')
        if not self.cluster_name:
            raise ValueError("No cluster_name found in {}".format(corosync.conf()))

    @qnetd_lock_for_multi_cluster
    def init_tls_certs_on_qnetd(self):
        """Initialize NSS database and generates CA and server certs on QNetd server."""
        cmd = "test -f {}".format(self.qnetd_cacert_on_qnetd)
        try:
            parallax.parallax_call([self.qnetd_addr], cmd)
            return
        except ValueError:
            # target file not exist
            pass

        logger.info('Generating QNetd CA and server certificates on %s', self.qnetd_addr)
        cmd = "corosync-qnetd-certutil -i"
        parallax.parallax_call([self.qnetd_addr], cmd)

    def fetch_qnetd_crt_from_qnetd(self, log: typing.Callable[[str, typing.Optional[str]], None]):
        """Fetch QNetd CA certificate(qnetd-cacert.crt) from QNetd server"""
        if os.path.exists(self.qnetd_cacert_on_local):
            return

        desc = "Fetch {} from {}".format(self.qnetd_cacert_filename, self.qnetd_addr)
        log(desc)
        crmsh.parallax.parallax_slurp([self.qnetd_addr], self.qdevice_path, self.qnetd_cacert_on_qnetd)

    def copy_qnetd_crt_to_cluster(self, log: typing.Callable[[str, typing.Optional[str]], None]):
        """Copy exported QNetd CA certificate (qnetd-cacert.crt) to every node"""
        node_list = utils.list_cluster_nodes_except_me()
        if not node_list:
            return

        desc = "Copy exported {} to {}".format(self.qnetd_cacert_filename, node_list)
        log(desc)
        self._copy_file_to_remote_hosts(
            os.path.dirname(self.qnetd_cacert_on_local),
            node_list, self.qdevice_path,
            recursive=True,
        )

    @staticmethod
    def _enclose_inet6_addr(addr: str):
        if ':' in addr:
            return f'[{addr}]'
        else:
            return addr

    @classmethod
    def _copy_file_to_remote_hosts(cls, local_file, remote_hosts: typing.Iterable[str], remote_path, recursive=False):
        crmsh.parallax.parallax_copy(remote_hosts, local_file, remote_path, recursive)

    def init_db_on_cluster(self, log: typing.Callable[[str, typing.Optional[str]], None]):
        """
        On one of cluster node initialize database by running
        /usr/sbin/corosync-qdevice-net-certutil -i -c qnetd-cacert.crt
        """
        node_list = utils.list_cluster_nodes()
        cmd = "corosync-qdevice-net-certutil -i -c {}".format(self.qnetd_cacert_on_local)
        desc = "Initialize database on {}".format(node_list)
        log(desc, cmd)
        crmsh.parallax.parallax_call(node_list, cmd)

    def create_ca_request(self, log: typing.Callable[[str, typing.Optional[str]], None]):
        """Generate certificate request:
        /usr/sbin/corosync-qdevice-net-certutil -r -n Cluster
        (Cluster name must match cluster_name key in the corosync.conf)
        """
        cmd = "corosync-qdevice-net-certutil -r -n {}".format(self.cluster_name)
        log("Generate certificate request {}".format(self.qdevice_crq_filename), cmd)
        sh.cluster_shell().get_stdout_or_raise_error(cmd)

    def copy_crq_to_qnetd(self, log: typing.Callable[[str, typing.Optional[str]], None]):
        """Copy exported CRQ to QNetd server"""
        desc = "Copy {} to {}".format(self.qdevice_crq_filename, self.qnetd_addr)
        log(desc)
        self._copy_file_to_remote_hosts(self.qdevice_crq_on_local, [self.qnetd_addr], self.qdevice_crq_on_qnetd)

    def sign_crq_on_qnetd(self, log: typing.Callable[[str, typing.Optional[str]], None]):
        """On QNetd server sign and export cluster certificate by running
        corosync-qnetd-certutil -s -c qdevice-net-node.crq -n Cluster
        """
        desc = "Sign and export cluster certificate on {}".format(self.qnetd_addr)
        cmd = "corosync-qnetd-certutil -s -c {} -n {}".\
                format(self.qdevice_crq_on_qnetd, self.cluster_name)
        log(desc, cmd)
        parallax.parallax_call([self.qnetd_addr], cmd)

    def fetch_cluster_crt_from_qnetd(self, log: typing.Callable[[str, typing.Optional[str]], None]):
        """Copy exported CRT to node where certificate request was created"""
        desc = "Fetch {} from {}".format(os.path.basename(self.qnetd_cluster_crt_on_qnetd), self.qnetd_addr)
        log(desc)
        crmsh.parallax.parallax_slurp([self.qnetd_addr], self.qdevice_path, self.qnetd_cluster_crt_on_qnetd)

    def import_cluster_crt(self, log: typing.Callable[[str, typing.Optional[str]], None]):
        """Import certificate on node where certificate request was created by
        running /usr/sbin/corosync-qdevice-net-certutil -M -c cluster-Cluster.crt
        """
        cmd = "corosync-qdevice-net-certutil -M -c {}".format(self.qnetd_cluster_crt_on_local)
        log("Import certificate file {} on local".format(os.path.basename(self.qnetd_cluster_crt_on_local)), cmd)
        sh.cluster_shell().get_stdout_or_raise_error(cmd)

    def copy_p12_to_cluster(self, log: typing.Callable[[str, typing.Optional[str]], None]):
        """Copy output qdevice-net-node.p12 to all other cluster nodes"""
        node_list = utils.list_cluster_nodes_except_me()
        if not node_list:
            return

        desc = "Copy {} to {}".format(self.qdevice_p12_filename, node_list)
        log(desc)
        self._copy_file_to_remote_hosts(self.qdevice_p12_on_local, node_list, self.qdevice_p12_on_local)

    def import_p12_on_cluster(self, log: typing.Callable[[str, typing.Optional[str]], None]):
        """Import cluster certificate and key on all other cluster nodes:
        /usr/sbin/corosync-qdevice-net-certutil -m -c qdevice-net-node.p12
        """
        node_list = utils.list_cluster_nodes_except_me()
        if not node_list:
            return

        desc = "Import {} on {}".format(self.qdevice_p12_filename, node_list)
        cmd = "corosync-qdevice-net-certutil -m -c {}".format(self.qdevice_p12_on_local)
        log(desc, cmd)
        QDevice.log_only_to_file(desc, cmd)
        parallax.parallax_call(node_list, cmd)

    def certificate_process_on_init(self):
        """
        The qdevice certificate process on init node
        """
        for i, step in enumerate([
            self.fetch_qnetd_crt_from_qnetd,
            self.copy_qnetd_crt_to_cluster,
            self.init_db_on_cluster,
            self.create_ca_request,
            self.copy_crq_to_qnetd,
            self.sign_crq_on_qnetd,
            self.fetch_cluster_crt_from_qnetd,
            self.import_cluster_crt,
            self.copy_p12_to_cluster,
            self.import_p12_on_cluster,
        ]):
            step(lambda s, cmd=None: self.log_only_to_file(f'Step {i+1}: {s}', cmd))

    def fetch_qnetd_crt_from_cluster(self):
        """
        Certificate process for join
        Step 1
        Fetch QNetd CA certificate(qnetd-cacert.crt) from init node
        """
        if os.path.exists(self.qnetd_cacert_on_cluster):
            return

        desc = "Step 1: Fetch {} from {}".format(self.qnetd_cacert_filename, self.cluster_node)
        QDevice.log_only_to_file(desc)
        crmsh.parallax.parallax_slurp([self.cluster_node], self.qdevice_path, self.qnetd_cacert_on_local)

    def init_db_on_local(self):
        """
        Certificate process for join
        Step 2
        Initialize database by running
        /usr/sbin/corosync-qdevice-net-certutil -i -c qnetd-cacert.crt
        """
        if os.path.exists(self.qdevice_db_path):
            utils.rmdir_r(self.qdevice_db_path)

        cmd = "corosync-qdevice-net-certutil -i -c {}".format(self.qnetd_cacert_on_cluster)
        QDevice.log_only_to_file("Step 2: Initialize database on local", cmd)
        sh.cluster_shell().get_stdout_or_raise_error(cmd)

    def fetch_p12_from_cluster(self):
        """
        Certificate process for join
        Step 3
        Fetch p12 key file from init node
        """
        if os.path.exists(self.qdevice_p12_on_cluster):
            return

        desc = "Step 3: Fetch {} from {}".format(self.qdevice_p12_filename, self.cluster_node)
        QDevice.log_only_to_file(desc)
        crmsh.parallax.parallax_slurp([self.cluster_node], self.qdevice_path, self.qdevice_p12_on_local)

    def import_p12_on_local(self):
        """
        Certificate process for join
        Step 4
        Import cluster certificate and key
        """
        cmd = "corosync-qdevice-net-certutil -m -c {}".format(self.qdevice_p12_on_cluster)
        QDevice.log_only_to_file("Step 4: Import cluster certificate and key", cmd)
        sh.cluster_shell().get_stdout_or_raise_error(cmd)

    def certificate_process_on_join(self):
        """
        The qdevice certificate process on join node
        """
        self.fetch_qnetd_crt_from_cluster()
        self.init_db_on_local()
        self.fetch_p12_from_cluster()
        self.import_p12_on_local()

    def write_qdevice_config(self) -> None:
        """
        Write qdevice attributes to config file
        """
        inst = corosync.ConfParser()
        qdevice_config_dict = {
                "model": "net",
                "net": {
                    "tls": self.tls,
                    "host": self.qnetd_addr,
                    "port": self.port,
                    "algorithm": self.algo,
                    "tie_breaker": self.tie_breaker
                    }
                }
        if self.algo == "ffsplit":
            # According to man corosync-qdevice, if the algorithm is lms, do not set votes
            qdevice_config_dict['votes'] = 1
        inst.set("quorum.device", qdevice_config_dict)
        if self.cmds:
            heuristics_dict = {"mode": self.mode}
            for i, cmd in enumerate(self.cmds.strip(';').split(';')):
                cmd_name = re.sub("[.-]", "_", os.path.basename(cmd.split()[0]))
                exec_name = "exec_{}{}".format(cmd_name, i)
                heuristics_dict[exec_name] = cmd
            inst.set("quorum.device.heuristics", heuristics_dict)
        inst.save()

    @staticmethod
    def remove_qdevice_config():
        """
        Remove configuration of qdevice
        """
        inst = corosync.ConfParser()
        inst.remove("quorum.device")
        inst.save()

    @staticmethod
    def remove_qdevice_db(addr_list=[]):
        """
        Remove qdevice database
        """
        if not os.path.exists(QDevice.qdevice_db_path):
            return
        node_list = addr_list if addr_list else utils.list_cluster_nodes()
        cmd = "rm -rf {}/*".format(QDevice.qdevice_path)
        QDevice.log_only_to_file("Remove qdevice database", cmd)
        parallax.parallax_call(node_list, cmd)

    @classmethod
    def remove_certification_files_on_qnetd(cls) -> None:
        """
        Remove this cluster related .crq and .crt files on qnetd
        """
        if not corosync.is_qdevice_configured():
            return
        qnetd_host = corosync.get_value('quorum.device.net.host')
        cluster_name = corosync.get_value('totem.cluster_name')
        cls_inst = cls(qnetd_host, cluster_name=cluster_name)
        shell = sh.cluster_shell()
        cmd = "test -f {crt_file} && rm -f {crt_file}".format(crt_file=cls_inst.qnetd_cluster_crt_on_qnetd)
        shell.get_stdout_or_raise_error(cmd, qnetd_host)
        cmd = "test -f {crq_file} && rm -f {crq_file}".format(crq_file=cls_inst.qdevice_crq_on_qnetd)
        shell.get_stdout_or_raise_error(cmd, qnetd_host)

    def config_qdevice(self) -> None:
        """
        Update configuration and reload corosync if necessary
        """
        self.write_qdevice_config()
        with logger_utils.status_long("Update configuration"):
            corosync.configure_two_node(qdevice_adding=True)
            bootstrap.sync_file(corosync.conf())
            if self.qdevice_reload_policy == QdevicePolicy.QDEVICE_RELOAD:
                sh.cluster_shell().get_stdout_or_raise_error("corosync-cfgtool -R")

    def config_qnetd_port(self):
        """
        Enable qnetd port in firewalld
        """
        if not ServiceManager().service_is_active("firewalld.service", self.qnetd_addr):
            return
        if utils.check_port_open(self.qnetd_addr, self.port):
            return
        shell = sh.cluster_shell()
        cmd = f"firewall-cmd --add-port={self.port}/tcp --permanent"
        rc, out, err = shell.get_rc_stdout_stderr_without_input(self.qnetd_addr, cmd)
        if rc != 0 and err:
            logger.error("Failed to add port {} to firewalld on {}: {}".format(self.port, self.qnetd_addr, err))
            return
        logger.info("Add port {} to firewalld on {}".format(self.port, self.qnetd_addr))
        shell.get_stdout_or_raise_error("firewall-cmd --reload", self.qnetd_addr)

    def start_qdevice_service(self):
        """
        Start qdevice and qnetd service
        """
        logger.info("Enable corosync-qdevice.service in cluster")
        utils.cluster_run_cmd("systemctl enable corosync-qdevice")
        if self.qdevice_reload_policy == QdevicePolicy.QDEVICE_RELOAD:
            logger.info("Starting corosync-qdevice.service in cluster")
            utils.cluster_run_cmd("systemctl restart corosync-qdevice")
        elif self.qdevice_reload_policy == QdevicePolicy.QDEVICE_RESTART:
            bootstrap.restart_cluster()
        else:
            logger.warning("To use qdevice service, need to restart cluster service manually on each node")

        logger.info("Enable corosync-qnetd.service on {}".format(self.qnetd_addr))
        self.enable_qnetd()
        logger.info("Starting corosync-qnetd.service on {}".format(self.qnetd_addr))
        self.start_qnetd()

    def adjust_sbd_watchdog_timeout_with_qdevice(self):
        """
        Adjust SBD_WATCHDOG_TIMEOUT when configuring qdevice and diskless SBD
        """
        self.using_diskless_sbd = SBDUtils.is_using_diskless_sbd()
        # add qdevice after diskless sbd started
        if self.using_diskless_sbd:
            res = SBDUtils.get_sbd_value_from_config("SBD_WATCHDOG_TIMEOUT")
            if not res or int(res) < SBDTimeout.SBD_WATCHDOG_TIMEOUT_DEFAULT_WITH_QDEVICE:
                sbd_watchdog_timeout_qdevice = SBDTimeout.SBD_WATCHDOG_TIMEOUT_DEFAULT_WITH_QDEVICE
                SBDManager.update_sbd_configuration({"SBD_WATCHDOG_TIMEOUT": str(sbd_watchdog_timeout_qdevice)})
                utils.set_property("stonith-timeout", SBDTimeout.get_stonith_timeout())

    @qnetd_lock_for_same_cluster_name
    def config_and_start_qdevice(self):
        """
        Wrap function to collect functions to config and start qdevice
        """
        QDevice.remove_qdevice_db()
        if self.tls == "on" or self.tls == 'required':
            with logger_utils.status_long("Qdevice certification process"):
                self.certificate_process_on_init()
        self.adjust_sbd_watchdog_timeout_with_qdevice()
        self.qdevice_reload_policy = evaluate_qdevice_quorum_effect(QDEVICE_ADD, self.using_diskless_sbd, self.is_stage)
        self.config_qdevice()
        self.config_qnetd_port()
        self.start_qdevice_service()

    @staticmethod
    def check_qdevice_vote():
        """
        Check if qdevice can contribute vote
        """
        out = sh.cluster_shell().get_stdout_or_raise_error("corosync-quorumtool -s", success_exit_status={0, 2})
        res = re.search(r'\s+0\s+0\s+Qdevice', out)
        if res:
            qnetd_host = corosync.get_value('quorum.device.net.host')
            logger.warning("Qdevice's vote is 0, which simply means Qdevice can't talk to Qnetd({}) for various reasons.".format(qnetd_host))

    @staticmethod
    def log_only_to_file(desc, cmd=None):
        logger_utils.log_only_to_file(desc)
        if cmd:
            logger_utils.log_only_to_file(f"Run: {cmd}")



"""parser and serializer for corosync.conf alike format"""


import logging
import re
from io import StringIO
import typing


logger = logging.getLogger(__name__)
COMMENT_PREFIX = '#comment'


class Parser:
    """SAX-style parser for the configuration of corosync"""

    _SPLIT_RE = re.compile('\\s*(#|{|:|})\\s*')

    def __init__(self, ifile):
        """Parse data form a file-like object."""
        self.__section_stack = list()
        self._parse(ifile)

    def enter_section(self, name: str):
        logger.info('entering: %s', name)

    def exit_section(self, name: str):
        logger.info('exiting: %s', name)

    def on_key_value(self, key: str, value: str):
        logger.info('%s: %s', key, value)

    def _parse(self, ifile):
        for lineno, line in enumerate(ifile):
            line = line.strip()
            if not line:
                continue
            try:
                tokens = self._tokenize(line)
            except ValueError as e:
                raise MalformedLineException(lineno, line)
            logger.debug('tokens: %s', tokens)
            if tokens[1] == '#':
                fake_key = f'{COMMENT_PREFIX}_{lineno}'
                self.on_key_value(fake_key, tokens[2])
            elif tokens[1] == '{':
                if not tokens[0] or tokens[2]:
                    raise MalformedLineException(lineno, line)
                self.__section_stack.append(tokens[0])
                self.enter_section(tokens[0])
            elif tokens[1] == ':':
                if not tokens[0]:
                    raise MalformedLineException(lineno, line)
                self.on_key_value(tokens[0], tokens[2])
            else:   # if tokens[1] == '}'
                if tokens[0] or tokens[2]:
                    raise MalformedLineException(lineno, line)
                if self.__section_stack:
                    self.exit_section(self.__section_stack[-1])
                    del self.__section_stack[-1]
                else:
                    raise UnbalancedBraceException(lineno + 1)
        if self.__section_stack:
            raise UnbalancedBraceException(0)

    @classmethod
    def _tokenize(cls, line):
        split_match = cls._SPLIT_RE.search(line)
        if split_match is None:
            raise ValueError()
        else:
            return (line[:split_match.start()], split_match.group(1), line[split_match.end():])


class ParserException(Exception):
    pass


class MalformedLineException(ParserException):
    def __init__(self, lineno, text):
        super().__init__(f'Malformed line {lineno}: {text}')
        self.lineno = lineno
        self.text = text


class UnbalancedBraceException(ParserException):
    def __init__(self, lineno):
        super().__init__(f'Unbalanced brace at line {lineno}')
        self.lineno = lineno


class DomParser(Parser):
    """parse a config stream into DOM tree"""
    def __init__(self, ifile):
        self._root = dict()
        self.__node_stack = [self._root]
        super().__init__(ifile)

    def dom(self):
        return self._root

    def __current_node(self):
        return self.__node_stack[-1]

    def enter_section(self, name: str):
        if name not in self.__current_node():
            new_node = dict()
            self.__current_node()[name] = new_node
            self.__node_stack.append(new_node)
            return
        node = self.__current_node()[name]
        new_node = dict()
        match node:
            case list(_):
                node.append(new_node)
            case _:
                self.__current_node()[name] = [node, new_node]
        self.__node_stack.append(new_node)

    def on_key_value(self, key: str, value: str):
        if key not in self.__current_node():
            self.__current_node()[key] = value
            return
        node = self.__current_node()[key]
        match node:
            case list(_):
                node.append(value)
            case _:
                self.__current_node()[key] = [node, value]

    def exit_section(self, name: str):
        del self.__node_stack[-1]


class DomQuery:
    """run queries on an DOM tree"""

    def __init__(self, dom):
        self._dom = dom

    def get(self, path: str | typing.List[str], index=0):
        """
        Gets the value for the path

        path: config path
        index: known index in section
        """
        if isinstance(path, str):
            return self.get(path.split('.'), index)
        node = self._dom
        for key in path:
            match node:
                case dict(_):
                    node = node[key]
                case _:
                    raise KeyError(path)
            while isinstance(node, list):
                node = node[index]
        return node

    def get_all(self, path: str | typing.List[str]):
        """
        Returns all values matching path
        """
        if isinstance(path, str):
            return self.get_all(path.split('.'))
        node = self._dom
        for i, key in enumerate(path):
            match node:
                case dict(_):
                    node = node[key]
                case _:
                    raise KeyError(path)
            if isinstance(node, list):
                result = list()
                for item in node:
                    result.extend(DomQuery(item).get_all(path[i+1:]))
                return result
        return [node]

    def enumerate_all_paths(self):
        queue = [(self._dom, tuple())]
        result = set()
        while queue:
            node, path = queue.pop(0)
            match node:
                case dict(_):
                    for key, value in node.items():
                        if key.startswith(COMMENT_PREFIX):
                            continue
                        queue.append((value, (*path, key)))
                case list(_) as li:
                    for item in li:
                        queue.append((item, path))
                case _:
                    result.add(path)
        return ['.'.join(item) for item in result]

    def remove(self, path, index=0):
        path = path.split('.')
        node = self.get(path[:-1], index)  # get parent node
        if not isinstance(node, dict):
            raise KeyError(path)
        key = path[-1]
        value = node[path[-1]]
        match value:
            case list(_):
                del value[index]
            case _:
                del node[key]


class DomSerializer:
    """serialize a DOM into bytes"""
    def __init__(self, node, ofile, indent='\t'):
        self._ofile = ofile
        self._indent = indent
        self._path_stack = list()
        match node:
            case dict(_):
                self.on_dict(node)
            case list(_):
                raise TypeError('invalid to serialize a list')
            case _:
                raise TypeError('invalid to serialize a scalar value')

    def on_dict(self, node):
        for key in self._sort_dict_keys(node.keys()):
            value = node[key]
            match value:
                case dict(_):
                    self.__write_indent(len(self._path_stack))
                    self._ofile.write(key)
                    self._ofile.write(' {\n')
                    self._path_stack.append(key)
                    self.on_dict(value)
                    del self._path_stack[-1]
                    self.__write_indent(len(self._path_stack))
                    self._ofile.write('}\n\n')
                case list(_):
                    self._path_stack.append(key)
                    self.on_list(value)
                    del self._path_stack[-1]
                case _:
                    self.__write_indent(len(self._path_stack))
                    if key.startswith(COMMENT_PREFIX):
                        self._ofile.write('# ')
                    else:
                        self._ofile.write(key)
                        self._ofile.write(': ')
                    self.on_value(value)
                    self._ofile.write('\n')

    def on_list(self, node):
        key = self._path_stack[-1]
        for item in node:
            match item:
                case dict(_):
                    self.__write_indent(len(self._path_stack) - 1)
                    self._ofile.write(key)
                    self._ofile.write(' {\n')
                    self.on_dict(item)
                    self.__write_indent(len(self._path_stack) - 1)
                    self._ofile.write('}\n\n')
                case list(_):
                    raise ValueError('list of list is invalid')
                case _:
                    self.__write_indent(len(self._path_stack) - 1)
                    self._ofile.write(key)
                    self._ofile.write(': ')
                    self.on_value(item)
                    self._ofile.write('\n')

    def on_value(self, value):
        self._ofile.write(str(value))

    @classmethod
    def _sort_dict_keys(cls, li):
        ret = list()
        first_ringx_addr_index = -1
        for i, x in enumerate(li):
            if re.match('^ring._addr$', x):
                if first_ringx_addr_index == -1:
                    # keep all ringX_addr placed together
                    first_ringx_addr_index = i
                ret.append((first_ringx_addr_index, x))
            else:
                ret.append((i, x))
        ret.sort()
        for _, x in ret:
            yield x

    def __write_indent(self, n):
        for _ in range(n):
            self._ofile.write(self._indent)



#!/usr/bin/python3
# Copyright (C) 2013 Kristoffer Gronlund <kgronlund@suse.com>
# See COPYING for license information.

import os
import sys
import subprocess
import json


DRY_RUN = False


def get_platform():
    return os.uname()[0]


def fail(msg):
    print(msg, file=sys.stderr)
    sys.exit(1)


def run(cmd):
    proc = subprocess.Popen(cmd,
                            shell=False,
                            stdin=None,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)
    out, err = proc.communicate(None)
    return proc.returncode, out, err


def is_program(prog):
    """Is this program available?"""
    for p in os.getenv("PATH").split(os.pathsep):
        filename = os.path.join(p, prog)
        if os.path.isfile(filename) and os.access(filename, os.X_OK):
            return filename
    return None


class PackageManager(object):
    def dispatch(self, name, state):
        if state in ('installed', 'present'):
            return self.present(name)
        elif state in ('absent', 'removed'):
            return self.absent(name)
        elif state == 'latest':
            return self.latest(name)
        fail(msg="Unknown state: " + state)

    def present(self, name):
        raise NotImplementedError

    def latest(self, name):
        raise NotImplementedError

    def absent(self, name):
        raise NotImplementedError


class Zypper(PackageManager):
    def __init__(self):
        self._rpm = is_program('rpm')
        self._zyp = is_program('zypper')
        if self._rpm is None or self._zyp is None:
            raise OSError("Missing tools: %s, %s" % (self._rpm, self._zyp))

    def get_version(self, name):
        cmd = [self._rpm, '-q', name]
        rc, stdout, stderr = run(cmd)
        if rc == 0:
            for line in stdout.splitlines():
                if name in line.decode('utf-8'):
                    return line.strip()
        return None

    def is_installed(self, name):
        if not isinstance(self._rpm, str):
            raise IOError(str(self._rpm))
        if not isinstance(name, str):
            raise IOError(str(name))
        cmd = [self._rpm, '--query', '--info', name]
        rc, stdout, stderr = run(cmd)
        return rc == 0

    def present(self, name):
        if self.is_installed(name):
            return (0, b'', b'', False)

        if DRY_RUN:
            return (0, b'', b'', True)

        cmd = [self._zyp,
               '--non-interactive',
               '--no-refresh',
               'install',
               '--auto-agree-with-licenses',
               name]
        rc, stdout, stderr = run(cmd)
        changed = rc == 0
        return (rc, stdout, stderr, changed)

    def latest(self, name):
        if not self.is_installed(name):
            return self.present(name)

        if DRY_RUN:
            return (0, b'', b'', True)

        pre_version = self.get_version(name)
        cmd = [self._zyp,
               '--non-interactive',
               '--no-refresh',
               'update',
               '--auto-agree-with-licenses',
               name]
        rc, stdout, stderr = run(cmd)
        post_version = self.get_version(name)
        changed = pre_version != post_version
        return (rc, stdout, stderr, changed)

    def absent(self, name):
        if not self.is_installed(name):
            return (0, b'', b'', False)

        if DRY_RUN:
            return (0, b'', b'', True)

        cmd = [self._zyp,
               '--non-interactive',
               'remove',
               name]
        rc, stdout, stderr = run(cmd)
        changed = rc == 0
        return (rc, stdout, stderr, changed)


class Yum(PackageManager):
    def __init__(self):
        self._rpm = is_program('rpm')
        self._yum = is_program('yum')

    def get_version(self, name):
        cmd = [self._rpm, '-q', name]
        rc, stdout, stderr = run(cmd)
        if rc == 0:
            for line in stdout.splitlines():
                if name in line:
                    return line.strip()
        return None

    def is_installed(self, name):
        cmd = [self._rpm, '--query', '--info', name]
        rc, stdout, stderr = run(cmd)
        return rc == 0

    def present(self, name):
        if self.is_installed(name):
            return (0, b'', b'', False)

        if DRY_RUN:
            return (0, b'', b'', True)

        cmd = [self._yum,
               '--assumeyes',
               '-d', '2',
               'install',
               name]
        rc, stdout, stderr = run(cmd)
        changed = rc == 0
        return (rc, stdout, stderr, changed)

    def latest(self, name):
        if not self.is_installed(name):
            return self.present(name)

        if DRY_RUN:
            return (0, b'', b'', True)

        pre_version = self.get_version(name)
        cmd = [self._yum,
               '--assumeyes',
               '-d', '2',
               'update',
               name]
        rc, stdout, stderr = run(cmd)
        post_version = self.get_version(name)
        changed = pre_version != post_version
        return (rc, stdout, stderr, changed)

    def absent(self, name):
        if not self.is_installed(name):
            return (0, b'', b'', False)

        if DRY_RUN:
            return (0, b'', b'', True)

        cmd = [self._yum,
               '--assumeyes',
               '-d', '2',
               'erase',
               name]
        rc, stdout, stderr = run(cmd)
        changed = rc == 0
        return (rc, stdout, stderr, changed)


class Apt(PackageManager):
    def __init__(self):
        self._apt = is_program('apt-get')
        self._dpkg = is_program('dpkg')
        if self._apt is None or self._dpkg is None:
            raise OSError("Missing tools: %s, %s" % (self._apt, self._dpkg))

    def get_version(self, name):
        cmd = [self._dpkg, '--status', name]
        rc, stdout, stderr = run(cmd)
        if rc == 0:
            for line in stdout.splitlines():
                info = line.decode('utf-8').split(':', 1)
                if len(info) == 2 and info[0] == 'Version':
                    return info[1].strip()
        return None

    def is_installed(self, name):
        cmd = [self._dpkg, '--status', name]
        rc, stdout, stderr = run(cmd)
        if rc == 0:
            for line in stdout.splitlines():
                info = line.decode('utf-8').split(':', 1)
                if len(info) == 2 and info[0] == 'Status':
                    return info[1].strip().endswith('installed')
        return False

    def present(self, name):
        if self.is_installed(name):
            return (0, b'', b'', False)

        if DRY_RUN:
            return (0, b'', b'', True)

        cmd = [self._apt,
               '--assume-yes',
               '--quiet',
               'install',
               name]
        rc, stdout, stderr = run(cmd)
        changed = rc == 0
        return (rc, stdout, stderr, changed)

    def latest(self, name):
        if not self.is_installed(name):
            return self.present(name)

        if DRY_RUN:
            return (0, b'', b'', True)

        pre_version = self.get_version(name)
        cmd = [self._apt,
               '--assume-yes',
               '--quiet',
               '--only-upgrade',
               'install',
               name]
        rc, stdout, stderr = run(cmd)
        post_version = self.get_version(name)
        changed = pre_version != post_version
        return (rc, stdout, stderr, changed)

    def absent(self, name):
        if not self.is_installed(name):
            return (0, b'', b'', False)

        if DRY_RUN:
            return (0, b'', b'', True)

        cmd = [self._apt,
               '--assume-yes',
               '--quiet',
               'purge',
               name]
        rc, stdout, stderr = run(cmd)
        changed = rc == 0
        return (rc, stdout, stderr, changed)


class Pacman(PackageManager):
    pass


def manage_package(pkg, state):
    """
    Gathers version and release information about a package.
    """
    if pkg is None:
        raise IOError("PKG IS NONE")
    pf = get_platform()
    if pf != 'Linux':
        fail(msg="Unsupported platform: " + pf)
    managers = {
        'zypper': Zypper,
        'yum': Yum,
        'apt-get': Apt,
        #'pacman': Pacman
    }
    for name, mgr in managers.items():
        exe = is_program(name)
        if exe:
            rc, stdout, stderr, changed = mgr().dispatch(pkg, state)
            return {'rc': rc,
                    'stdout': stdout.decode('utf-8'),
                    'stderr': stderr.decode('utf-8'),
                    'changed': changed
                    }
    fail(msg="No supported package manager found")


def main():
    import argparse
    parser = argparse.ArgumentParser(
        description="(Semi)-Universal package installer",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('-d', '--dry-run', dest='dry_run', action='store_true',
                        help="Only check if changes would be made")

    parser.add_argument('-n', '--name', metavar='name', type=str,
                        help="Name of package")

    parser.add_argument('-s', '--state', metavar='state', type=str,
                        help="Desired state (present|latest|removed)", default="present")

    args = parser.parse_args()
    global DRY_RUN
    DRY_RUN = args.dry_run
    if not args.name or not args.state:
        raise IOError("Bad arguments: %s" % (sys.argv))
    data = manage_package(args.name, args.state)
    print(json.dumps(str(data)))

main()



# Copyright (C) 2008-2011 Dejan Muhamedagic <dmuhamedagic@suse.de>
# See COPYING for license information.
cib_cli_map = {
    "node": "node",
    "primitive": "primitive",
    "group": "group",
    "clone": "clone",
    "master": "ms",
    "bundle": "bundle",
    "rsc_location": "location",
    "rsc_colocation": "colocation",
    "rsc_order": "order",
    "rsc_ticket": "rsc_ticket",
    "template": "rsc_template",
    "cluster_property_set": "property",
    "rsc_defaults": "rsc_defaults",
    "op_defaults": "op_defaults",
    "acl_target": "acl_target",
    "acl_group": "acl_group",
    "acl_user": "user",
    "acl_role": "role",
    "fencing-topology": "fencing_topology",
    "tag": "tag",
    "alert": "alert",
}
container_tags = ("group", "clone", "ms", "master", "bundle")
clonems_tags = ("clone", "ms", "master")
resource_tags = ("primitive", "group", "clone", "ms", "master", "template", "bundle")
constraint_tags = ("rsc_location", "rsc_colocation", "rsc_order", "rsc_ticket")
constraint_rsc_refs = ("rsc", "with-rsc", "first", "then")
children_tags = ("group", "primitive")
nvpairs_tags = ("meta_attributes", "instance_attributes", "utilization")
defaults_tags = ("rsc_defaults", "op_defaults")
resource_cli_names = ("primitive", "group", "clone", "ms", "master", "rsc_template", "bundle")
constraint_cli_names = ("location", "colocation", "collocation", "order", "rsc_ticket")
nvset_cli_names = ("property", "rsc_defaults", "op_defaults")
op_cli_names = ("monitor",
                "start",
                "stop",
                "migrate_to",
                "migrate_from",
                "promote",
                "demote",
                "notify",
                "reload")
ra_operations = tuple(["probe"] + list(op_cli_names))

subpfx_list = {
    "instance_attributes": "instance_attributes",
    "meta_attributes": "meta_attributes",
    "utilization": "utilization",
    "operations": "ops",
    "rule": "rule",
    "expression": "expression",
    "date_expression": "expression",
    "duration": "duration",
    "date_spec": "date_spec",
    "read": "read",
    "write": "write",
    "deny": "deny",
}
acl_rule_names = ("read", "write", "deny")
acl_spec_map = {
    "xpath": "xpath",
    "ref": "ref",
    "tag": "tag",
    "attribute": "attribute",
}
# ACLs were rewritten in pacemaker 1.1.12
# this is the new acl syntax
acl_spec_map_2 = {
    "xpath": "xpath",
    "ref": "reference",
    "reference": "reference",
    "tag": "object-type",
    "type": "object-type",
    "attr": "attribute",
    "attribute": "attribute"
}

acl_spec_map_2_rev = (('xpath', 'xpath'),
                      ('reference', 'ref'),
                      ('attribute', 'attr'),
                      ('object-type', 'type'))

acl_shortcuts = {
    "meta":
    (r"//primitive\[@id='@@'\]/meta_attributes", r"/nvpair\[@name='@@'\]"),
    "params":
    (r"//primitive\[@id='@@'\]/instance_attributes", r"/nvpair\[@name='@@'\]"),
    "utilization":
    (r"//primitive\[@id='@@'\]/utilization",),
    "location":
    (r"//rsc_location\[@id='cli-prefer-@@' and @rsc='@@'\]",),
    "property":
    (r"//crm_config/cluster_property_set", r"/nvpair\[@name='@@'\]"),
    "nodeattr":
    (r"//nodes/node/instance_attributes", r"/nvpair\[@name='@@'\]"),
    "nodeutil":
    (r"//nodes/node/utilization", r"\[@uname='@@'\]"),
    "node":
    (r"//nodes/node", r"\[@uname='@@'\]"),
    "status":
    (r"/cib/status",),
    "cib":
    (r"/cib",),
}
lrm_exit_codes = {
    "success": "0",
    "unknown": "1",
    "args": "2",
    "unimplemented": "3",
    "perm": "4",
    "installed": "5",
    "configured": "6",
    "not_running": "7",
    "promoted": "8",
    "failed_promoted": "9",
}
lrm_status_codes = {
    "pending": "-1",
    "done": "0",
    "cancelled": "1",
    "timeout": "2",
    "notsupported": "3",
    "error": "4",
}
cib_user_attrs = ("validate-with",)
node_states = ("online", "offline", "unclean")
precious_attrs = ("id-ref",)
op_extra_attrs = ("interval",)
rsc_meta_attributes = (
    "allow-migrate", "maintenance", "is-managed", "interval-origin",
    "migration-threshold", "priority", "multiple-active",
    "failure-timeout", "resource-stickiness", "target-role",
    "description", "remote-node", "requires", "provides",
    "remote-port", "remote-addr", "remote-connect-timeout",
    "critical", "allow-unhealthy-nodes", "container-attribute-target"
)
common_meta_attributes = ("priority", "target-role", "is-managed")
group_meta_attributes = common_meta_attributes + ("container", )
clone_meta_attributes = common_meta_attributes + (
    "ordered", "notify", "interleave", "globally-unique",
    "clone-max", "clone-node-max", "clone-state", "description",
    "clone-min", "promotable", "promoted-max", "promoted-node-max",
)
ms_meta_attributes = common_meta_attributes + (
    "clone-max", "clone-node-max", "notify", "globally-unique", "ordered", 
    "interleave", "master-max", "master-node-max", "description",
)
bundle_meta_attributes = common_meta_attributes
alert_meta_attributes = (
    "timeout", "timestamp-format"
)
trace_ra_attr = "trace_ra"
trace_dir_attr = "trace_dir"
score_types = {'advisory': '0', 'mandatory': 'INFINITY'}
boolean_ops = ('or', 'and')
binary_ops = ('lt', 'gt', 'lte', 'gte', 'eq', 'ne')
binary_types = ('string', 'version', 'number')
unary_ops = ('defined', 'not_defined')
simple_date_ops = ('lt', 'gt')
date_ops = ('lt', 'gt', 'in_range', 'date_spec')
date_spec_names = '''hours monthdays weekdays yearsdays months \
weeks years weekyears'''.split()
in_range_attrs = ('start', 'end')
node_default_type = "member"
node_attributes_keyw = ("attributes", "utilization")
shadow_envvar = "CIB_shadow"
attr_defaults = {
    "node": {"type": "member"},
    "resource_set": {"sequential": "true", "require-all": "true"},
    "rule": {"boolean-op": "and"},
}
cib_no_section_rc = 6
# Graphviz attributes for various CIB elements.
# Shared for edge and node and graph attributes.
# Keys are graphviz attributes, values are dicts where keys
# are CIB element names and values graphviz values.
# - element "." refers to the whole graph
# - element "class:<ra_class>" refers to primitives of a
#   specific RA class
# - optional_set is a resource_set with require-all set to
#   false
# - group and optional_set are subgraphs (boxes)
graph = {
    ".": {
        "compound": "true",
    },
    "*": {
        "fontname": "Helvetica",
        "fontsize": "11",
    },
    "node": {
        "style": "bold",
        "shape": "box",
        "color": "#7ac142",
    },
    "primitive": {
        "fillcolor": "#e4e5e6",
        "color": "#b9b9b9",
        "shape": "box",
        "style": "rounded,filled",
    },
    "rsc_template": {
        "fillcolor": "#ffd457",
        "color": "#b9b9b9",
        "shape": "box",
        "style": "rounded,filled,dashed",
    },
    "class:stonith": {
        "shape": "box",
        "style": "dashed",
    },
    "location": {
        "style": "dashed",
        "dir": "none",
    },
    "clone": {
        "color": "#ec008c",
    },
    "ms": {
        "color": "#f8981d",
    },
    "bundle": {
        "color": "#00aeef",
        "style": "rounded",
    },
    "group": {
        "color": "#00aeef",
        "group": "#00aeef",
        "labelloc": "b",
        "labeljust": "r",
        "labelfontsize": "12",
    },
    "optional_set": {
        "style": "dotted",
    },
    "template:edge": {
        "color": "#b9b9b9",
        "style": "dotted",
        "arrowtail": "open",
        "dir": "back",
    },
}

need_reset = False
prompt = ''
tmp_cib = False
tmp_cib_prompt = "@tmp@"
live_cib_prompt = "live"

simulate_programs = {
    "ptest": "ptest",
    "simulate": "crm_simulate",
}

meta_progs_20 = ("pacemaker-controld", "pacemaker-schedulerd", "pacemaker-fenced", "pacemaker-based")

# elide these properties from tab completion
controld_metadata_do_not_complete = ("dc-version", "cluster-infrastructure")
extra_cluster_properties = ("dc-version",
                            "cluster-infrastructure",
                            "last-lrm-refresh",
                            "cluster-name")
pcmk_version = ""  # set later

container_type = ("docker", "podman")
container_helptxt = {
    "container": {
        "image": """image:(string)
    Container image tag(required)""",

        "replicas": """replicas:(integer)
    Default:Value of promoted-max if that is positive, else 1
    A positive integer specifying the number of container instances to launch""",

        "replicas-per-host": """replicas-per-host:(integer)
    Default:1
    A positive integer specifying the number of container instances allowed to
    run on a single node""",

        "promoted-max": """promoted-max:(integer)
    Default:0
    A non-negative integer that, if positive, indicates that the containerized
    service should be treated as a promotable service, with this many replicas
    allowed to run the service in the promoted role""",

        "run-command": """run-command:(string)
    Default:/usr/sbin/pacemaker_remoted if bundle contains a primitive, otherwise none
    This command will be run inside the container when launching it ("PID 1").
    If the bundle contains a primitive, this command must start pacemaker_remoted
    (but could, for example, be a script that does other stuff, too).""",

        "options": """options:(string)
    Extra command-line options to pass to the 'docker run' or 'podman run' command"""
    },

    "network": {
        "add-host": """add-host:(string)
    Default:True
    If True, and ip-range-start is specified, Pacemake will automatically ensure that
    /etc/hosts inside the containers has entries for each replica name and its assigned IP.""",

        "ip-range-start": """ip-range-start:(IPv4 address)
    If specified, Pacemaker will create an implicit ocf:heartbeat:IPaddr2 resource
    for each container instance, starting with this IP address, using up to replicas
    sequential addresses. These addresses can be used from the host’s network to
    reach the service inside the container, though it is not visible within the
    container itself. Only IPv4 addresses are currently supported.""",

        "host-netmask": """host-netmask:(integer)
    Default:32
    If ip-range-start is specified, the IP addresses are created with this CIDR
    netmask (as a number of bits).""",

        "host-interface": """host-interface:(string)
    If ip-range-start is specified, the IP addresses are created on this host
    interface (by default, it will be determined from the IP address).""",

        "control-port": """control-port:(integer)
    Default: 3121
    If the bundle contains a primitive, the cluster will use this integer TCP port
    for communication with Pacemaker Remote inside the container. Changing this is
    useful when the container is unable to listen on the default port, for example,
    when the container uses the host’s network rather than ip-range-start (in which
    case replicas-per-host must be 1), or when the bundle may run on a Pacemaker
    Remote node that is already listening on the default port. Any PCMK_remote_port
    environment variable set on the host or in the container is ignored for bundle
    connections.""",

        "port-mapping": {
            "id": """id:(string)
    A unique name for the port mapping (required)""",

            "port": """port:(integer)
    If this is specified, connections to this TCP port number on the host network
    (on the container’s assigned IP address, if ip-range-start is specified) will
    be forwarded to the container network. Exactly one of port or range must be
    specified in a port-mapping.""",

            "internal-port": """internal-port:(integer)
    Default: value of port
    If port and this are specified, connections to port on the host’s network will
    be forwarded to this port on the container network.""",

                "range": """range:(first_port-last_port)
    If this is specified, connections to these TCP port numbers (expressed as
    first_port-last_port) on the host network (on the container’s assigned IP address,
    if ip-range-start is specified) will be forwarded to the same ports in the container
    network. Exactly one of port or range must be specified in a port-mapping."""
        }
    },

    "storage": {
        "id": """id:(string)
    A unique name for the storage mapping (required)""",

        "source-dir": """source-dir:(string)
    The absolute path on the host’s filesystem that will be mapped into the container.
    Exactly one of source-dir and source-dir-root must be specified in a storage-mapping.""",

        "source-dir-root": """source-dir-root:(string)
    The start of a path on the host’s filesystem that will be mapped into the container,
    using a different subdirectory on the host for each container instance. The subdirectory
    will be named the same as the replica name. Exactly one of source-dir and source-dir-root
    must be specified in a storage-mapping.""",

           "target-dir": """target-dir:(string)
    The path name within the container where the host storage will be mapped (required)""",

            "options": """options:(string)
    A comma-separated list of file system mount options to use when mapping the storage"""
    },
}


QDEVICE_HELP_INFO = """  QDevice participates in quorum decisions. With the assistance of 
  a third-party arbitrator Qnetd, it provides votes so that a cluster 
  is able to sustain more node failures than standard quorum rules 
  allow. It is recommended for clusters with an even number of nodes 
  and highly recommended for 2 node clusters."""


SSH_OPTION_ARGS = ["-o", "StrictHostKeyChecking=no"]
SSH_OPTION = ' '.join(SSH_OPTION_ARGS)


CLOUD_AWS = "amazon-web-services"
CLOUD_AZURE = "microsoft-azure"
CLOUD_GCP = "google-cloud-platform"


RED = '\033[31m'
YELLOW = '\033[33m'
GREEN = '\033[32m'
END = '\033[0m'


CIB_QUERY = "cibadmin -Q"
CIB_UPGRADE = "crm configure upgrade force"
CIB_RAW_FILE = "/var/lib/pacemaker/cib/cib.xml"
XML_NODE_PATH = "/cib/configuration/nodes/node"
XML_STATUS_PATH = "/cib/status/node_state"
XML_NODE_QUERY_STANDBY_PATH = "//nodes/node[@id='{node_id}']/instance_attributes/nvpair[@name='standby']"
XML_STATUS_QUERY_STANDBY_PATH = "//status/node_state[@id='{node_id}']/transient_attributes/instance_attributes/nvpair[@name='standby']"
CRM_MON_ONE_SHOT = "crm_mon -1"
CRM_MON_XML_OUTPUT= "crm_mon --output-as=xml"
STONITH_TIMEOUT_DEFAULT = 60
PCMK_DELAY_MAX = 30
DLM_CONTROLD_RA = "ocf:pacemaker:controld"
LVMLOCKD_RA = "ocf:heartbeat:lvmlockd"
HA_USER = "hacluster"
HA_GROUP = "haclient"
SCHEMA_MIN_VER_SUPPORT_OCF_1_1 = "pacemaker-3.7"
REJOIN_COUNT = 60
REJOIN_INTERVAL = 10
DC_DEADTIME_DEFAULT = 20

ADVISED_ACTION_LIST = ['monitor', 'start', 'stop', 'promote', 'demote']
ADVISED_KEY_LIST = ['timeout', 'interval', 'role']
DEFAULT_INTERVAL_IN_ACTION = "20s"
WAIT_TIMEOUT_MS_DEFAULT = 120000
CSYNC2_SERVICE = "csync2.socket"

RSC_ROLE_PROMOTED = "Promoted"
RSC_ROLE_UNPROMOTED = "Unpromoted"
RSC_ROLE_PROMOTED_LEGACY = "Master"
RSC_ROLE_UNPROMOTED_LEGACY = "Slave"
PCMK_VERSION_DEFAULT = "2.0.0"

INTERFACE_HELP = """Bind to IP address on interface IF. 
Allowed value is nic name or IP address. 
If a nic name is provided, the first IP of that nic will be used. 
Use multiple -i for more links. Note: Only one link is allowed for the non knet transport type
"""

COROSYNC_STATUS_TYPES = ("ring", "quorum", "qdevice", "qnetd", "cpg")

NO_SSH_ERROR_MSG = "ssh-related operations are disabled. crmsh works in local mode."

PCMK_SERVICE = "pacemaker.service"
SBD_SERVICE = "sbd.service"
# vim:ts=4:sw=4:et:



# Copyright (C) 2020 Xin Liang <XLiang@suse.com>
# See COPYING for license information.


import re
import time
from contextlib import contextmanager

from . import sh
from . import config
from . import log
from .sh import ShellUtils


logger = log.setup_logger(__name__)


class SSHError(Exception):
    """
    Custom exception for ssh error
    """


class ClaimLockError(Exception):
    """
    Custom exception if claiming lock failed or wait lock release timed out
    """


class Lock(object):
    """
    A base class define a lock mechanism used to exclude other nodes
    """

    LOCK_DIR_DEFAULT = "/run/.crmsh_lock_directory"

    def __init__(self, lock_dir=None):
        """
        Init function
        """
        # only the lock owner can unlock
        self.lock_owner = False
        self.lock_dir = lock_dir or self.LOCK_DIR_DEFAULT

    def _run(self, cmd):
        """
        Run command on local
        """
        return ShellUtils().get_stdout_stderr(cmd)

    def _create_lock_dir(self):
        """
        Create lock directory, mkdir command was atomic
        """
        cmd = "mkdir {}".format(self.lock_dir)
        rc, _, _ = self._run(cmd)
        if rc == 0:
            self.lock_owner = True
            return True
        return False

    def _lock_or_fail(self):
        """
        Raise ClaimLockError if claiming lock failed
        """
        if not self._create_lock_dir():
            raise ClaimLockError("Failed to claim lock (the lock directory exists at {})".format(self.lock_dir))

    def _unlock(self):
        """
        Remove the lock directory
        """
        if self.lock_owner:
            cmd = "rm -rf {}".format(self.lock_dir)
            self._run(cmd)

    @contextmanager
    def lock(self):
        """
        Create lock directory on local, and remove it finally
        Might raise ClaimLockError
        """
        try:
            self._lock_or_fail()
            yield
        except:
            raise
        finally:
            self._unlock()


class RemoteLock(Lock):
    """
    A class inherited from Lock class
    Define the behavior how to claim lock on remote node and how to wait the lock released
    """

    SSH_TIMEOUT = 10
    SSH_OPTION = "-o ConnectTimeout={} -o StrictHostKeyChecking=no".format(SSH_TIMEOUT)
    SSH_EXIT_ERR = 255
    MIN_LOCK_TIMEOUT = 120
    WAIT_INTERVAL = 10

    def __init__(self, remote_node, for_join=True, lock_dir=None, wait=True, no_warn=False):
        """
        Init function
        """
        self.remote_node = remote_node
        self.for_join = for_join
        self.wait = wait
        self.no_warn = no_warn
        super(__class__, self).__init__(lock_dir=lock_dir)

    def _run(self, cmd):
        """
        Run command on remote node
        """
        # TODO: pass SSH_OPTION
        rc, out, err = sh.cluster_shell().get_rc_stdout_stderr_without_input(self.remote_node, cmd)
        if rc == self.SSH_EXIT_ERR:
            raise SSHError(err)
        return rc, out, err

    @property
    def lock_timeout(self):
        """
        Get lock_timeout from config.core
        """
        try:
            value = int(config.core.lock_timeout)
        except ValueError:
            raise ValueError("Invalid format of core.lock_timeout(should be a number)")
        if value < self.MIN_LOCK_TIMEOUT:
            raise ValueError("Minimum value of core.lock_timeout should be {}".format(self.MIN_LOCK_TIMEOUT))
        return value

    def _get_online_nodelist(self):
        """
        Get the online node list from remote node
        """
        rc, out, err = self._run("crm_node -l")
        if rc != 0 and err:
            raise ValueError(err)
        return re.findall('[0-9]+ (.*) member', out)

    def _lock_or_wait(self):
        """
        Try to claim lock on remote node, wait if failed to claim
        raise ClaimLockError if reached the lock_timeout
        """
        warned_once = False
        online_list = []
        pre_online_list = []
        expired_error_str = "Cannot continue since the lock directory exists at the node ({}:{})".format(self.remote_node, self.lock_dir)

        current_time = int(time.time())
        timeout = current_time + self.lock_timeout
        while current_time <= timeout:

            # Try to claim the lock
            if self._create_lock_dir():
                # Success
                break

            if self.for_join:
                # Might lose claiming lock again, start to wait again
                online_list = self._get_online_nodelist()
                if pre_online_list and pre_online_list != online_list:
                    timeout = current_time + self.lock_timeout
                pre_online_list = online_list

            if not self.no_warn and not warned_once:
                warned_once = True
                logger.warning("Might have unfinished process on other nodes, wait %ss...", self.lock_timeout)

            time.sleep(self.WAIT_INTERVAL)
            current_time = int(time.time())

        else:
            raise ClaimLockError("Timed out after {} seconds. {}".format(self.lock_timeout, expired_error_str))

    @contextmanager
    def lock(self):
        """
        Create lock directory on remote, and remove it finally
        Might raise SSHError, ClaimLockError and ValueError
        """
        try:
            if self.wait:
                self._lock_or_wait()
            else:
                self._lock_or_fail()
            yield
        except:
            raise
        finally:
            self._unlock()



# Copyright (C) 2008-2011 Dejan Muhamedagic <dmuhamedagic@suse.de>
# Copyright (C) 2013-2018 Kristoffer Gronlund <kgronlund@suse.com>
# See COPYING for license information.

import re
from . import command
from . import completers as compl
from . import constants
from . import config
from . import utils
from . import xmlutil
from . import ui_utils
from . import options
from .cibconfig import cib_factory
from .sh import ShellUtils
from . import log


logger = log.setup_logger(__name__)
logger_utils = log.LoggerUtils(logger)


def rm_meta_attribute(node, attr, l, force_children=False):
    '''
    Build a list of nvpair nodes which contain attribute
    (recursively in all children resources)
    '''
    for c in node.iterchildren():
        if c.tag == "meta_attributes":
            nvpair = xmlutil.get_attr_in_set(c, attr)
            if nvpair is not None:
                l.append(nvpair)
        elif force_children or \
                (xmlutil.is_child_rsc(c) and not c.getparent().tag == "group"):
            rm_meta_attribute(c, attr, l, force_children=force_children)

def get_children_with_attr(node, attr):
    l = []
    for p in node.xpath(".//primitive"):
        diff_attr = False
        for meta_set in xmlutil.get_set_nodes(p, "meta_attributes", create=False):
            p_value = xmlutil.get_attr_value(meta_set, attr)
            if p_value is not None:
                diff_attr = True
                break
        if diff_attr:
            l.append(p)
    return l

def set_deep_meta_attr_node(target_node, attr, value):
    nvpair_l = []
    conflicting_attr = ''
    if 'maintenance' == attr:
        conflicting_attr = 'is-managed'
    if 'is-managed' == attr:
        conflicting_attr = 'maintenance'
    if xmlutil.is_clone(target_node):
        for c in target_node.iterchildren():
            if xmlutil.is_child_rsc(c):
                rm_meta_attribute(c, attr, nvpair_l)
    if config.core.manage_children != "never" and \
            (xmlutil.is_group(target_node) or
             (xmlutil.is_clone(target_node) and xmlutil.cloned_el(target_node) == "group")):
        # maintenance and is-managed attributes conflict,
        # ask the user if he wants to leave both.
        if '' != conflicting_attr:
            odd_children = get_children_with_attr(target_node, conflicting_attr)
            for c in odd_children:
                if config.core.manage_children == "always" or \
                        (config.core.manage_children == "ask" and
                        utils.ask("'%s' conflicts with '%s' attribute. Remove '%s' for child resource %s?" %
                                (attr, conflicting_attr, conflicting_attr, c.get("id")))):
                    logger.debug("force remove meta attr %s from %s", conflicting_attr, c.get("id"))
                    rm_meta_attribute(c, conflicting_attr, nvpair_l, force_children=True)
        # remove the same attributes from all children (if wanted)
        odd_children = get_children_with_attr(target_node, attr)
        for c in odd_children:
            if config.core.manage_children == "always" or \
                    (config.core.manage_children == "ask" and
                     utils.ask("Do you want to override '%s' for child resource %s?" %
                               (attr, c.get("id")))):
                logger.debug("force remove meta attr %s from %s", attr, c.get("id"))
                rm_meta_attribute(c, attr, nvpair_l, force_children=True)
    xmlutil.rmnodes(list(set(nvpair_l)))
    if '' != conflicting_attr:
        nvpairs = target_node.xpath("./meta_attributes/nvpair[@name='%s']" % (conflicting_attr))
        if len(nvpairs) > 0:
            if (utils.ask("'%s' conflicts with '%s' attribute. Remove '%s' for resource %s?" %
                        (attr, conflicting_attr, conflicting_attr, target_node.get("id")))):
                xmlutil.rmnodes(list(set(nvpairs)))
    xmlutil.xml_processnodes(target_node,
                             xmlutil.is_emptynvpairs, xmlutil.rmnodes)

    # work around issue with pcs interoperability
    # by finding exising nvpairs -- if there are any, just
    # set the value in those. Otherwise fall back to adding
    # to all meta_attributes tags
    nvpairs = target_node.xpath("./meta_attributes/nvpair[@name='%s']" % (attr))
    if len(nvpairs) > 0:
        for nvpair in nvpairs:
            nvpair.set("value", value)
    else:
        for n in xmlutil.get_set_nodes(target_node, "meta_attributes", create=True):
            xmlutil.set_attr(n, attr, value)
    return True


def set_deep_meta_attr(rsc, attr, value, commit=True):
    """
    If the referenced rsc is a primitive that belongs to a group,
    then set its attribute.
    Otherwise, go up to the topmost resource which contains this
    resource and set the attribute there (i.e. if the resource is
    cloned).
    If it's a group then check its children. If any of them has
    the attribute set to a value different from the one given,
    then ask the user whether to reset them or not (exact
    behaviour depends on the value of config.core.manage_children).
    """

    def update_obj(obj):
        """
        set the meta attribute in the given object
        """
        node = obj.node
        obj.set_updated()
        if not (node.tag == "primitive" and
                node.getparent().tag == "group"):
            node = xmlutil.get_topmost_rsc(node)
        return set_deep_meta_attr_node(node, attr, value)

    def flatten(objs):
        for obj in objs:
            if isinstance(obj, list):
                for subobj in obj:
                    yield subobj
            else:
                yield obj

    def resolve(obj):
        if obj.obj_type == 'tag':
            ret = [cib_factory.find_object(o) for o in obj.node.xpath('./obj_ref/@id')]
            ret = [r for r in ret if r is not None]
            return ret
        return obj

    def is_resource(obj):
        return xmlutil.is_resource(obj.node)

    objs = cib_factory.find_objects(rsc)
    if objs is None:
        logger.error("CIB is not valid!")
        return False
    while any(obj for obj in objs if obj.obj_type == 'tag'):
        objs = list(flatten(resolve(obj) for obj in objs))
    objs = list(filter(is_resource, objs))
    logger.debug("set_deep_meta_attr: %s", ', '.join([obj.obj_id for obj in objs]))
    if not objs:
        logger.error("Resource not found: %s", rsc)
        return False

    ok = all(update_obj(obj) for obj in objs)
    if not ok:
        logger.error("Failed to update meta attributes for %s", rsc)
        return False

    if not commit:
        return True

    ok = cib_factory.commit()
    if not ok:
        logger.error("Failed to commit updates to %s", rsc)
        return False
    return True


ATTR_OPERATION_CMDS = ('delete', 'set', 'show')
_attrcmds = compl.choice(ATTR_OPERATION_CMDS)
_raoperations = compl.choice(constants.ra_operations)


class RscMgmt(command.UI):
    '''
    Resources management class
    '''
    name = "resource"

    rsc_status_all = "crm_resource --list"
    rsc_status = "crm_resource --locate --resource '%s'"
    rsc_showxml = "crm_resource --query-xml --resource '%s'"
    rsc_setrole = "crm_resource --meta --resource '%s' --set-parameter target-role --parameter-value '%s'"
    rsc_migrate = "crm_resource --quiet --move --resource '%s' %s"
    rsc_unmigrate = "crm_resource --quiet --clear --resource '%s'"
    rsc_ban = "crm_resource --ban --resource '%s' %s"
    rsc_maintenance = "crm_resource --resource '%s' --meta --set-parameter maintenance --parameter-value '%s'"
    rsc_param = {
        'set': "crm_resource --resource '%s' --set-parameter '%s' --parameter-value '%s'",
        'delete': "crm_resource --resource '%s' --delete-parameter '%s'",
        'show': "crm_resource --resource '%s' --get-parameter '%s'",
        'get': "crm_resource --resource '%s' --get-parameter '%s'",
    }
    rsc_meta = {
        'set': "crm_resource --meta --resource '%s' --set-parameter '%s' --parameter-value '%s'",
        'delete': "crm_resource --meta --resource '%s' --delete-parameter '%s'",
        'show': "crm_resource --meta --resource '%s' --get-parameter '%s'",
        'get': "crm_resource --meta --resource '%s' --get-parameter '%s'",
    }
    rsc_failcount = {
        'set': "crm_attribute -t status -n 'fail-count-%s' -N '%s' -v '%s'",
        'set_p': "crm_attribute -t status -P 'fail-count-%s' -N '%s' -v '%s'",
        'delete': "crm_failcount -D -r %s -N %s",
        'show': "crm_failcount -G -r %s -N %s",
        'get': "crm_failcount -G -r %s -N %s",
    }
    rsc_utilization = {
        'set': "crm_resource --utilization --resource '%s' --set-parameter '%s' --parameter-value '%s'",
        'delete': "crm_resource --utilization --resource '%s' --delete-parameter '%s'",
        'show': "crm_resource --utilization --resource '%s' --get-parameter '%s'",
        'get': "crm_resource --utilization --resource '%s' --get-parameter '%s'",
    }
    rsc_secret = {
        'set': "cibsecret set '%s' '%s' '%s'",
        'stash': "cibsecret stash '%s' '%s'",
        'unstash': "cibsecret unstash '%s' '%s'",
        'delete': "cibsecret delete '%s' '%s'",
        'show': "cibsecret get '%s' '%s'",
        'get': "cibsecret get '%s' '%s'",
        'check': "cibsecret check '%s' '%s'",
    }

    def _refresh_cleanup(self, action, rsc, node, force):
        """
        Implements the refresh and cleanup commands.
        """
        if rsc == "force":
            rsc, force = None, True
        if node == "force":
            node, force = None, True
        cmd = ["crm_resource", "--" + action]
        if rsc:
            if not utils.is_name_sane(rsc):
                return False
            cmd.append("--resource")
            cmd.append(rsc)
        if node:
            if not utils.is_name_sane(node):
                return False
            cmd.append("--node")
            cmd.append(node)
        if force:
            cmd.append("--force")
        return utils.ext_cmd(" ".join(cmd)) == 0

    def requires(self):
        for program in ('crm_resource', 'crm_attribute'):
            if not utils.is_program(program):
                logger_utils.no_prog_err(program)
                return False
        return True

    @command.alias('show', 'list')
    @command.completers_repeating(compl.resources)
    def do_status(self, context, *resources):
        "usage: status [<rsc> ...]"
        if len(resources) > 0:
            rc = True
            for rsc in resources:
                if not utils.is_name_sane(rsc):
                    return False
                rc = rc and (utils.ext_cmd(self.rsc_status % rsc) == 0)
            return rc
        else:
            return utils.ext_cmd(self.rsc_status_all) == 0

    def _commit_meta_attr(self, context, rsc, name, value):
        """
        Perform change to resource
        """
        if not utils.is_name_sane(rsc):
            return False
        commit = not cib_factory.has_cib_changed()
        if not commit:
            context.info("Currently editing the CIB, changes will not be committed")
        return set_deep_meta_attr(rsc, name, value, commit=commit)

    def _commit_meta_attrs(self, context, resources, name, value):
        """
        Perform change to list of resources
        """
        for rsc in resources:
            if not utils.is_name_sane(rsc):
                return False
        commit = not cib_factory.has_cib_changed()
        if not commit:
            context.info("Currently editing the CIB, changes will not be committed")

        rc = True
        for rsc in resources:
            rc = rc and set_deep_meta_attr(rsc, name, value, commit=False)
        if commit and rc:
            ok = cib_factory.commit()
            if not ok:
                logger.error("Failed to commit updates to %s", rsc)
            return ok
        return rc

    @command.wait
    @command.completers_repeating(compl.resources)
    def do_start(self, context, *resources):
        "usage: start <rsc> [<rsc> ...]"
        if len(resources) == 0:
            context.error("Expected at least one resource as argument")
        return self._commit_meta_attrs(context, resources, "target-role", "Started")

    @command.wait
    @command.completers_repeating(compl.resources)
    def do_stop(self, context, *resources):
        "usage: stop <rsc> [<rsc> ...]"
        if len(resources) == 0:
            context.error("Expected at least one resource as argument")
        return self._commit_meta_attrs(context, resources, "target-role", "Stopped")

    @command.wait
    @command.completers_repeating(compl.resources)
    def do_restart(self, context, *resources):
        "usage: restart <rsc> [<rsc> ...]"
        logger.info("ordering %s to stop", ", ".join(resources))
        if not self._commit_meta_attrs(context, resources, "target-role", "Stopped"):
            return False
        if not utils.wait_dc_stable("stop", not options.batch):
            return False
        logger.info("ordering %s to start", ", ".join(resources))
        return self._commit_meta_attrs(context, resources, "target-role", "Started")

    @command.wait
    @command.completers(compl.resources)
    def do_promote(self, context, rsc):
        "usage: promote <rsc>"
        if not utils.is_name_sane(rsc):
            return False
        if not xmlutil.RscState().is_ms_or_promotable_clone(rsc):
            logger.error("%s is not a promotable resource", rsc)
            return False
        role = utils.handle_role_for_ocf_1_1(constants.RSC_ROLE_PROMOTED)
        return utils.ext_cmd(self.rsc_setrole % (rsc, role)) == 0

    def do_scores(self, context):
        "usage: scores"
        if utils.is_program('crm_simulate'):
            utils.ext_cmd('crm_simulate -sUL')
        elif utils.is_program('ptest'):
            utils.ext_cmd('ptest -sUL')
        else:
            context.fatal_error("Need crm_simulate or ptest in path to display scores")

    @command.completers_repeating(compl.resources)
    def do_locate(self, context, *resources):
        "usage: locate <rsc> [<rsc> ...]"
        if len(resources) == 0:
            context.error("Expected at least one resource as argument")
        for rsc in resources:
            utils.ext_cmd("crm_resource --resource '%s' --locate" % (rsc))

    @command.wait
    @command.completers(compl.resources)
    def do_demote(self, context, rsc):
        "usage: demote <rsc>"
        if not utils.is_name_sane(rsc):
            return False
        if not xmlutil.RscState().is_ms_or_promotable_clone(rsc):
            logger.error("%s is not a promotable resource", rsc)
            return False
        role = utils.handle_role_for_ocf_1_1(constants.RSC_ROLE_UNPROMOTED)
        return utils.ext_cmd(self.rsc_setrole % (rsc, role)) == 0

    @command.completers(compl.resources)
    def do_manage(self, context, rsc):
        "usage: manage <rsc>"
        return self._commit_meta_attr(context, rsc, "is-managed", "true")

    @command.completers(compl.resources)
    def do_unmanage(self, context, rsc):
        "usage: unmanage <rsc>"
        return self._commit_meta_attr(context, rsc, "is-managed", "false")

    def move_or_ban(self, context, rsc, *args):
        """
        Common codes for move or ban action
        """
        if not utils.is_name_sane(rsc):
            return False
        cmd_map_dict = {'move': self.rsc_migrate,
                'ban': self.rsc_ban}
        action = context.get_command_name()
        action_cap = action.capitalize()
        action_cmd = cmd_map_dict[action]
        usage = "usage: {} <rsc> [<node>] [<lifetime>] [force]".format(action)
        node = None
        lifetime = None

        argl = list(args)
        force = "force" in utils.fetch_opts(argl, ["force"]) or config.core.force
        if len(argl) >= 3:
            context.fatal_error(usage)
        if len(argl) == 2:  # must be <node> <lifetime>
            node = argl[0]
            if not xmlutil.is_our_node(node):
                context.fatal_error("Not our node: " + node)
            lifetime = utils.fetch_lifetime_opt(argl)
        elif len(argl) == 1: # could be <node> or <lifetime>
            if xmlutil.is_our_node(argl[0]):
                node = argl[0]
            else:
                lifetime = utils.fetch_lifetime_opt(argl)

        if action == "move" and not node and not force:
            context.fatal_error("No target node: {} requires either a target node or 'force'".format(action_cap))

        opts = ''
        if node:
            opts = "--node '%s'" % node
        if lifetime:
            opts = "%s --lifetime '%s'" % (opts, lifetime)
        if force or config.core.force:
            opts = "%s --force" % opts
        rc = utils.ext_cmd(action_cmd % (rsc, opts))
        if rc == 0:
            if node:
                logger.info("%s constraint created for %s to %s", action_cap, rsc, node)
            else:
                logger.info("%s constraint created for %s", action_cap, rsc)
            logger.info("Use `crm resource clear %s` to remove this constraint", rsc)
        return rc == 0

    @command.alias('migrate')
    @command.skill_level('administrator')
    @command.wait
    @command.completers_repeating(compl.resources, compl.nodes)
    def do_move(self, context, rsc, *args):
        """usage: move <rsc> [<node>] [<lifetime>] [force]"""
        return self.move_or_ban(context, rsc, *args)


    @command.skill_level('administrator')
    @command.wait
    @command.completers_repeating(compl.resources, compl.nodes)
    def do_ban(self, context, rsc, *args):
        """usage: ban <rsc> [<node>] [<lifetime>] [force]"""
        return self.move_or_ban(context, rsc, *args)

    @command.alias('unmove', 'unban', 'unmigrate')
    @command.skill_level('administrator')
    @command.wait
    @command.completers(compl.resources)
    def do_clear(self, context, rsc):
        "usage: clear <rsc>"
        if not utils.is_name_sane(rsc):
            return False
        rc = utils.ext_cmd(self.rsc_unmigrate % rsc)
        if rc == 0:
            logger.info("Removed migration constraints for %s", rsc)
        return rc == 0

    @command.skill_level('administrator')
    @command.wait
    @command.completers(compl.resources, compl.nodes)
    def do_cleanup(self, context, rsc=None, node=None, force=False):
        "usage: cleanup [<rsc>] [<node>] [force]"
        return self._refresh_cleanup("cleanup", rsc, node, force)

    @command.wait
    @command.completers(compl.resources, compl.nodes)
    def do_operations(self, context, resource=None, node=None):
        "usage: operations [<rsc>] [<node>]"
        cmd = "crm_resource -O"
        if resource is None:
            return utils.ext_cmd(cmd)
        if node is None:
            return utils.ext_cmd("%s -r '%s'" % (cmd, resource))
        return utils.ext_cmd("%s -r '%s' -N '%s'" % (cmd, resource, node))

    @command.wait
    @command.completers(compl.resources)
    def do_constraints(self, context, resource):
        "usage: constraints <rsc>"
        return utils.ext_cmd("crm_resource -a -r '%s'" % (resource))

    @command.wait
    @command.completers(compl.resources, _attrcmds, compl.nodes)
    def do_failcount(self, context, rsc, cmd, node, value=None, operation=None, interval=None):
        """usage:
        failcount <rsc> set <node> <value> [operation] [interval]
        failcount <rsc> delete <node>
        failcount <rsc> show <node>"""
        def sec_to_ms(s):
            return '0' if s == '0' else s + '000'

        def ms_to_sec(m):
            return '0' if m == '0' else str(int(m) // 1000)

        USAGE = "Usage: failcount <rsc> set <node> <value> [operation] [interval]"

        if rsc not in compl.resources():
            logger.warning("Resource %s not in this cluster", rsc)
            return
        if cmd not in ATTR_OPERATION_CMDS:
            context.fatal_error(f"{cmd} is not valid command(should be one of {', '.join(ATTR_OPERATION_CMDS)})")
        nodeid = utils.get_nodeid_from_name(node)
        if nodeid is None:
            context.fatal_error(f"Node {node} not in this cluster")
        if value and (not value.isdigit() and value != "INFINITY"):
            context.fatal_error("value should be a non-negative integer or 'INFINITY'")

        if cmd == "set":
            # query the current failcount status
            query_cmd = "cibadmin -Q --xpath '/cib/status/node_state[@id='{}']'".format(nodeid)
            rc, out, err = ShellUtils().get_stdout_stderr(query_cmd)
            if rc != 0:
                context.fatal_error(err)

            # try to get failcount dict {operation:interval}
            failcount_dict = dict(re.findall(rf'fail-count-{rsc}#(.*)_([0-9]+)', out))
            if failcount_dict:
                # validate for operation and interval
                if operation and operation not in failcount_dict:
                    context.fatal_error(f"Should specify operation between ({', '.join(failcount_dict.keys())})\n{USAGE}")
                if (operation and interval) and (operation, sec_to_ms(interval)) not in failcount_dict.items():
                    failcount_items = [(op, ms_to_sec(inter)) for op, inter in failcount_dict.items()]
                    context.fatal_error(f"Should specify (operation, interval) between {failcount_items}\n{USAGE}")
            else:
                # no failcount entry for this resource and node
                if operation and interval:
                    rsc = f'{rsc}#{operation}_{sec_to_ms(interval)}'
                else:
                    # when no failcount entry, it's no harm to give a warning and return true
                    # https://bugzilla.suse.com/show_bug.cgi?id=1180332#c11
                    logger.warning("Should specify operation and interval\n%s", USAGE)
                    return

            # just one failcount entry
            if len(failcount_dict) == 1:
                operation, interval_in_ms = failcount_dict.popitem()
                rsc = f'{rsc}#{operation}_{interval_in_ms}'

            # multiple failcount entries for this resource and node
            if len(failcount_dict) > 1:
                if operation and interval:
                    rsc = f'{rsc}#{operation}_{sec_to_ms(interval)}'
                elif int(value) == 0:
                    op_interval_list = [
                        f'{op}_{inter}'
                        for op, inter in failcount_dict.items()
                        if not operation or op == operation
                    ]
                    op_interval_str = '|'.join(op_interval_list)
                    rsc = f'{rsc}#{op_interval_str}'
                    # using '-P' option of 'crm_attribute' command, to match the pattern
                    cmd = "set_p"
                else:
                    # value != 0
                    if operation and len([op for op in failcount_dict if op == operation]) == 1:
                        rsc = f'{rsc}#{operation}_{failcount_dict[operation]}'
                    else:
                        failcount_items = [(op, ms_to_sec(inter)) for op, inter in failcount_dict.items()]
                        context.fatal_error(f"Should specify (operation, interval) between {failcount_items}\n{USAGE}")

        return ui_utils.manage_attr(context.get_command_name(), self.rsc_failcount,
                                    rsc, cmd, node, value)

    @command.skill_level('administrator')
    @command.wait
    @command.completers(compl.resources, _attrcmds)
    def do_param(self, context, rsc, cmd, param, value=None):
        """usage:
        param <rsc> set <param> <value>
        param <rsc> delete <param>
        param <rsc> show <param>"""
        return ui_utils.manage_attr(context.get_command_name(), self.rsc_param,
                                    rsc, cmd, param, value)

    @command.skill_level('administrator')
    @command.completers(compl.resources,
                        compl.choice(['set', 'stash', 'unstash', 'delete', 'show', 'check']))
    def do_secret(self, context, rsc, cmd, param, value=None):
        """usage:
        secret <rsc> set <param> <value>
        secret <rsc> stash <param>
        secret <rsc> unstash <param>
        secret <rsc> delete <param>
        secret <rsc> show <param>
        secret <rsc> check <param>"""
        return ui_utils.manage_attr(context.get_command_name(), self.rsc_secret,
                                    rsc, cmd, param, value)

    @command.skill_level('administrator')
    @command.wait
    @command.completers(compl.resources, _attrcmds)
    def do_meta(self, context, rsc, cmd, attr, value=None):
        """usage:
        meta <rsc> set <attr> <value>
        meta <rsc> delete <attr>
        meta <rsc> show <attr>"""
        return ui_utils.manage_attr(context.get_command_name(), self.rsc_meta,
                                    rsc, cmd, attr, value)

    @command.skill_level('administrator')
    @command.wait
    @command.completers(compl.resources, _attrcmds)
    def do_utilization(self, context, rsc, cmd, attr, value=None):
        """usage:
        utilization <rsc> set <attr> <value>
        utilization <rsc> delete <attr>
        utilization <rsc> show <attr>"""
        return ui_utils.manage_attr(context.get_command_name(), self.rsc_utilization,
                                    rsc, cmd, attr, value)

    @command.alias('reprobe')
    @command.completers(compl.resources, compl.nodes)
    def do_refresh(self, context, rsc=None, node=None, force=False):
        'usage: refresh [<rsc>] [<node>] [force]'
        return self._refresh_cleanup("refresh", rsc, node, force)

    @command.wait
    @command.completers(compl.resources, compl.choice(['on', 'off', 'true', 'false']))
    def do_maintenance(self, context, resource, on_off='true'):
        'usage: maintenance <resource> [on|off|true|false]'
        on_off = on_off.lower()
        if on_off not in ('on', 'true', 'off', 'false'):
            context.fatal_error("Expected <resource> [on|off|true|false]")
        elif on_off in ('on', 'true'):
            on_off = 'true'
        else:
            on_off = 'false'
        return self._commit_meta_attr(context, resource, "maintenance", on_off)

    def _get_trace_rsc(self, rsc_id):
        if not cib_factory.refresh():
            return None
        rsc = cib_factory.find_object(rsc_id)
        if not rsc:
            logger.error("resource %s does not exist", rsc_id)
            return None
        if rsc.obj_type != "primitive":
            logger.error("element %s is not a primitive resource", rsc_id)
            return None
        return rsc

    def _add_trace_op(self, rsc, op, interval, dir):
        from lxml import etree
        n = etree.Element('op')
        n.set('name', op)
        n.set('interval', interval)
        n.set(constants.trace_ra_attr, '1')
        if dir is not None:
            n.set(constants.trace_dir_attr, dir)
        return rsc.add_operation(n)

    def _trace_resource(self, context, rsc_id, rsc, dir):
        """Enable RA tracing for a specified resource."""
        op_nodes = rsc.node.xpath('operations/op')

        def trace(name):
            for o in op_nodes:
                if o.get('name') == name:
                    return
            if not self._add_trace_op(rsc, name, '0', dir):
                context.fatal_error("Failed to add trace for %s:%s" % (rsc_id, name))
        trace('start')
        trace('stop')
        if xmlutil.is_ms_or_promotable_clone(rsc.node):
            trace('promote')
            trace('demote')
        for op_node in op_nodes:
            op_node = rsc.set_op_attr(op_node, constants.trace_ra_attr, "1")
            if dir is not None:
                rsc.set_op_attr(op_node, constants.trace_dir_attr, dir)

    def _trace_op(self, context, rsc_id, rsc, op, dir):
        """Enable RA tracing for a specified operation."""
        op_nodes = rsc.node.xpath('operations/op[@name="%s"]' % (op))
        if not op_nodes:
            if op == 'monitor':
                context.fatal_error("No monitor operation configured for %s" % (rsc_id))
            if not self._add_trace_op(rsc, op, '0', dir):
                context.fatal_error("Failed to add trace for %s:%s" % (rsc_id, op))
        for op_node in op_nodes:
            op_node = rsc.set_op_attr(op_node, constants.trace_ra_attr, "1")
            if dir is not None:
                rsc.set_op_attr(op_node, constants.trace_dir_attr, dir)

    def _trace_op_interval(self, context, rsc_id, rsc, op, interval, dir):
        """Enable RA tracing for an operation with the exact interval."""
        op_node = xmlutil.find_operation(rsc.node, op, interval)
        if op_node is None and utils.crm_msec(interval) != 0:
            context.fatal_error("Operation %s with interval %s not found in %s" % (op, interval, rsc_id))
        if op_node is None:
            if not self._add_trace_op(rsc, op, interval, dir):
                context.fatal_error("Failed to add trace for %s:%s" % (rsc_id, op))
        else:
            op_node = rsc.set_op_attr(op_node, constants.trace_ra_attr, "1")
            if dir is not None:
                rsc.set_op_attr(op_node, constants.trace_dir_attr, dir)

    @command.completers(compl.primitives, _raoperations)
    def do_trace(self, context, rsc_id, *args):
        'usage: trace <rsc> [<op>] [<interval>] [<log-dir>]'
        usage='usage: trace <rsc> [<op>] [<interval>] [<log-dir>]'
        rsc = self._get_trace_rsc(rsc_id)
        if not rsc:
            return False
        if len(args) > 3:
            context.fatal_error(usage)

        op, interval, trace_dir = None, None, None
        for arg in args:
            if arg.startswith("/"):
                if trace_dir is not None:
                    context.fatal_error(usage)
                trace_dir = arg
            elif arg.isnumeric():
                if interval is not None:
                    context.fatal_error(usage)
                interval = arg
            else:
                if op is not None:
                    context.fatal_error(usage)
                op = arg
        trace_dir = trace_dir or f"{config.path.heartbeat_dir}/trace_ra"

        if op == "probe":
            op = "monitor"
            interval = interval or "0"
        if op is None:
            self._trace_resource(context, rsc_id, rsc, trace_dir)
        elif interval is None:
            self._trace_op(context, rsc_id, rsc, op, trace_dir)
        else:
            self._trace_op_interval(context, rsc_id, rsc, op, interval, trace_dir)

        if not cib_factory.commit():
            return False

        rsc_type = rsc.node.get("type")
        logger.info("Trace for %s%s is written to %s", rsc_id, ":"+op if op else "", f"{trace_dir}/{rsc_type}")
        if op and op != "monitor":
            logger.info("Trace set, restart %s to trace the %s operation", rsc_id, op)
        else:
            logger.info("Trace set, restart %s to trace non-monitor operations", rsc_id)
        return True

    def _remove_trace(self, rsc, op_node):
        logger.debug("op_node: %s", xmlutil.xml_tostring(op_node))
        op_node = rsc.del_op_attr(op_node, constants.trace_ra_attr)
        op_node = rsc.del_op_attr(op_node, constants.trace_dir_attr)
        if rsc.is_dummy_operation(op_node):
            rsc.del_operation(op_node)

    def _untrace_resource(self, context, rsc_id, rsc):
        """Disable RA tracing for a specified resource."""
        op_nodes = rsc.node.xpath(
            'operations/op[instance_attributes/nvpair[@name="%s"]]' %
            (constants.trace_ra_attr))
        for op_node in op_nodes:
            self._remove_trace(rsc, op_node)

    def _untrace_op(self, context, rsc_id, rsc, op):
        """Disable RA tracing for a specified operation."""
        op_nodes = rsc.node.xpath('operations/op[@name="%s"]' % (op))
        if not op_nodes:
            context.fatal_error("Operation %s not found in %s" % (op, rsc_id))
        for op_node in op_nodes:
            self._remove_trace(rsc, op_node)

    def _untrace_op_interval(self, context, rsc_id, rsc, op, interval):
        """Disable RA tracing for an operation with the exact interval."""
        op_node = xmlutil.find_operation(rsc.node, op, interval)
        if op_node is None:
            context.fatal_error(
                "Operation %s with interval %s not found in %s" %
                (op, interval, rsc_id))
        self._remove_trace(rsc, op_node)

    @command.completers(compl.primitives, _raoperations)
    def do_untrace(self, context, rsc_id, op=None, interval=None):
        'usage: untrace <rsc> [<op>] [<interval>]'
        rsc = self._get_trace_rsc(rsc_id)
        if not rsc:
            return False
        if op == "probe":
            op = "monitor"
            if interval is None:
                interval = "0"
        if op is None:
            self._untrace_resource(context, rsc_id, rsc)
        elif interval is None:
            self._untrace_op(context, rsc_id, rsc, op)
        else:
            self._untrace_op_interval(context, rsc_id, rsc, op, interval)
        if not cib_factory.commit():
            return False
        logger.info("Stop tracing %s%s", rsc_id, " for operation "+op if op else "")
        return True



import ast
import re
import time
import os
import datetime
import yaml

import behave
from behave import given, when, then
import behave_agent
from crmsh import corosync, userdir, bootstrap
from crmsh import utils as crmutils
from crmsh import sbd
from crmsh import ui_configure
from crmsh.sh import ShellUtils
from utils import check_cluster_state, check_service_state, online, run_command, me, \
                  run_command_local_or_remote, file_in_archive, \
                  assert_eq, is_unclean, assert_in
import const


def _parse_str(text):
    return ast.literal_eval(text)
_parse_str.pattern='"([^"]|\\")*?"'


behave.use_step_matcher("cfparse")
behave.register_type(str=_parse_str)


@when('Write multi lines to file "{f}" on "{addr}"')
def step_impl(context, f, addr):
    data_list = context.text.split('\n')
    for line in data_list:
        echo_option = " -n" if line == data_list[-1] else ""
        cmd = "echo{} \"{}\"|sudo tee -a {}".format(echo_option, line, f)
        if addr != me():
            sudoer = userdir.get_sudoer()
            user = f"{sudoer}@" if sudoer else ""
            cmd = f"ssh {user}{addr} '{cmd}'"
        run_command(context, cmd)


@given('Cluster service is "{state}" on "{addr}"')
def step_impl(context, state, addr):
    assert check_cluster_state(context, state, addr) is True


@given('Nodes [{nodes:str+}] are cleaned up')
def step_impl(context, nodes):
    run_command(context, 'crm resource cleanup || true')
    for node in nodes:
        # wait for ssh service
        for _ in range(10):
            rc, _, _ = ShellUtils().get_stdout_stderr('ssh {} true'.format(node))
            if rc == 0:
                break
            time.sleep(1)
        run_command_local_or_remote(context, "crm cluster stop {} || true".format(node), node)
        assert check_cluster_state(context, 'stopped', node) is True


@given('Service "{name}" is "{state}" on "{addr}"')
def step_impl(context, name, state, addr):
    assert check_service_state(context, name, state, addr) is True


@given('Has disk "{disk}" on "{addr}"')
def step_impl(context, disk, addr):
    _, out, _ = run_command_local_or_remote(context, "fdisk -l", addr)
    assert re.search(r'{} '.format(disk), out) is not None


@given('Online nodes are "{nodelist}"')
def step_impl(context, nodelist):
    assert online(context, nodelist) is True


@given('Run "{cmd}" OK')
def step_impl(context, cmd):
    rc, _, _ = run_command(context, cmd)
    assert rc == 0


@then('Run "{cmd}" OK')
def step_impl(context, cmd):
    rc, _, _ = run_command(context, cmd)
    assert rc == 0


@when('Run "{cmd}" OK')
def step_impl(context, cmd):
    rc, _, _ = run_command(context, cmd)
    assert rc == 0


@given('IP "{addr}" is belong to "{iface}"')
def step_impl(context, addr, iface):
    cmd = 'ip address show dev {}'.format(iface)
    res = re.search(r' {}/'.format(addr), run_command(context, cmd)[1])
    assert bool(res) is True


@given('Run "{cmd}" OK on "{addr}"')
def step_impl(context, cmd, addr):
    _, out, _ = run_command_local_or_remote(context, cmd, addr, True)

@when('Run "{cmd}" on "{addr}"')
def step_impl(context, cmd, addr):
    _, out, _ = run_command_local_or_remote(context, cmd, addr)


@then('Run "{cmd}" OK on "{addr}"')
def step_impl(context, cmd, addr):
    _, out, _ = run_command_local_or_remote(context, cmd, addr)


@then('Print stdout')
def step_impl(context):
    context.logger.info("\n{}".format(context.stdout))


@then('Print stderr')
def step_impl(context):
    context.logger.info("\n{}".format(context.stderr))


@then('No crmsh tracebacks')
def step_impl(context):
    if "Traceback (most recent call last):" in context.stderr and \
            re.search('File "/usr/lib/python.*/crmsh/', context.stderr):
        context.logger.info("\n{}".format(context.stderr))
        context.failed = True


@when('Try "{cmd}" on "{addr}"')
def step_impl(context, cmd, addr):
    run_command_local_or_remote(context, cmd, addr, exit_on_fail=False)


@when('Try "{cmd}"')
def step_impl(context, cmd):
    _, out, _ = run_command(context, cmd, exit_on_fail=False)


@when('Wait "{second}" seconds')
def step_impl(context, second):
    time.sleep(int(second))


@then('Got output "{msg}"')
def step_impl(context, msg):
    assert context.stdout == msg
    context.stdout = None


@then('Expected multiple lines')
def step_impl(context):
    assert context.stdout == context.text
    context.stdout = None


@then('Expected "{msg}" in stdout')
def step_impl(context, msg):
    assert_in(msg, context.stdout)
    context.stdout = None


@then('Expected "{msg}" in stderr')
def step_impl(context, msg):
    assert_in(msg, context.stderr)
    context.stderr = None


@then('Expect stdout contains snippets [{snippets:str+}].')
def step_impl(context, snippets):
    for snippet in snippets:
        assert_in(snippet, context.stdout)
    context.stdout = None


@then('Expected regex "{reg_str}" in stdout')
def step_impl(context, reg_str):
    res = re.search(reg_str, context.stdout)
    assert res is not None
    context.stdout = None


@then('Expected regex "{reg_str}" in stderr')
def step_impl(context, reg_str):
    assert context.stderr is not None and re.search(reg_str, context.stderr)
    context.stderr = None


@then('Expected return code is "{num}"')
def step_impl(context, num):
    assert context.return_code == int(num)


@then('Expected "{msg}" not in stdout')
def step_impl(context, msg):
    assert msg not in context.stdout
    context.stdout = None


@then('Expected "{msg}" not in stderr')
def step_impl(context, msg):
    assert context.stderr is None or msg not in context.stderr
    context.stderr = None


@then('Except "{msg}"')
def step_impl(context, msg):
    assert_in(msg, context.stderr)
    context.stderr = None


@then('Except multiple lines')
def step_impl(context):
    assert_in(context.text, context.stderr)
    context.stderr = None


@then('Expected multiple lines in output')
def step_impl(context):
    assert_in(context.text, context.stdout)
    context.stdout = None


@then('Except "{msg}" in stderr')
def step_impl(context, msg):
    assert_in(msg, context.stderr)
    context.stderr = None


@then('Cluster service is "{state}" on "{addr}"')
def step_impl(context, state, addr):
    assert check_cluster_state(context, state, addr) is True


@then('Service "{name}" is "{state}" on "{addr}"')
def step_impl(context, name, state, addr):
    assert check_service_state(context, name, state, addr) is True


@then('Online nodes are "{nodelist}"')
def step_impl(context, nodelist):
    assert online(context, nodelist) is True


@then('Remote online nodes are "{nodelist}"')
def step_impl(context, nodelist):
    _, out, _ = run_command(context, 'crm_mon -1|grep RemoteOnline:')
    for node in nodelist.split():
        assert node in out, "Node {} not found in {}".format(node, out)


@then('Node "{node}" is standby')
def step_impl(context, node):
    assert crmutils.is_standby(node) is True


@then('Node "{node}" is online')
def step_impl(context, node):
    assert crmutils.is_standby(node) is False


@then('IP "{addr}" is used by corosync on "{node}"')
def step_impl(context, addr, node):
    nodeid = crmutils.get_nodeid_from_name(node)
    _, out, _ = run_command_local_or_remote(context, f'corosync-cfgtool -a {nodeid}', node)
    assert addr in out.split()


@then('Cluster name is "{name}"')
def step_impl(context, name):
    _, out, _ = run_command(context, 'corosync-cmapctl -b totem.cluster_name')
    assert out.split()[-1] == name


@then('Cluster virtual IP is "{addr}"')
def step_impl(context, addr):
    _, out, _ = run_command(context, 'crm configure show|grep -A1 IPaddr2')
    res = re.search(r' ip={}'.format(addr), out)
    assert bool(res) is True


@then('Cluster is using "{transport_type}" transport mode')
def step_impl(context, transport_type):
    assert corosync.get_value('totem.transport') == transport_type
    _, out, _ = run_command(context, 'corosync-cfgtool -s')
    assert re.search(f'transport {transport_type}\n', out) is not None


@then('two_node in corosync.conf is "{number}"')
def step_impl(context, number):
    assert corosync.get_value('quorum.two_node') == number


@then('Show cluster status on "{addr}"')
def step_impl(context, addr):
    _, out, _ = run_command_local_or_remote(context, 'crm_mon -1', addr)
    if out:
        context.logger.info("\n{}".format(out))


@then('Show corosync ring status')
def step_impl(context):
    _, out, _ = run_command(context, 'crm corosync status ring')
    if out:
        context.logger.info("\n{}".format(out))


@then('Show crm configure')
def step_impl(context):
    _, out, _ = run_command(context, 'crm configure show')
    if out:
        context.logger.info("\n{}".format(out))


@then('Show status from qnetd')
def step_impl(context):
    _, out, _ = run_command(context, 'crm corosync status qnetd')
    if out:
        context.logger.info("\n{}".format(out))


@then('Show qdevice status')
def step_impl(context):
    _, out, _ = run_command(context, 'crm corosync status qdevice')
    if out:
        context.logger.info("\n{}".format(out))


@then('Show corosync qdevice configuration')
def step_impl(context):
    _, out, _ = run_command(context, "sed -n -e '/quorum/,/^}/ p' /etc/corosync/corosync.conf")
    if out:
        context.logger.info("\n{}".format(out))


@then('Resource "{res}" type "{res_type}" is "{state}"')
def step_impl(context, res, res_type, state):
    try_count = 0
    result = None
    while try_count < 20:
        time.sleep(1)
        _, out, _ = run_command(context, "crm_mon -1rR")
        if out:
            result = re.search(r'\s{}\s+.*:+{}\):\s+{} '.format(res, res_type, state), out)
            if not result:
                try_count += 1
            else:
                break
    assert result is not None


@then('Resource "{res}" failcount on "{node}" is "{number}"')
def step_impl(context, res, node, number):
    cmd = "crm resource failcount {} show {}".format(res, node)
    _, out, _ = run_command(context, cmd)
    if out:
        result = re.search(r'name=fail-count-{} value={}'.format(res, number), out)
        assert result is not None


@then('Resource "{res_type}" not configured')
def step_impl(context, res_type):
    _, out, _ = run_command(context, "crm configure show")
    result = re.search(r' {} '.format(res_type), out)
    assert result is None


@then('Output is the same with expected "{cmd}" help output')
def step_impl(context, cmd):
    cmd_help = {}
    cmd_help["crm"] = const.CRM_H_OUTPUT
    cmd_help["crm_cluster_init"] = const.CRM_CLUSTER_INIT_H_OUTPUT
    cmd_help["crm_cluster_join"] = const.CRM_CLUSTER_JOIN_H_OUTPUT
    cmd_help["crm_cluster_remove"] = const.CRM_CLUSTER_REMOVE_H_OUTPUT
    cmd_help["crm_cluster_geo-init"] = const.CRM_CLUSTER_GEO_INIT_H_OUTPUT
    cmd_help["crm_cluster_geo-join"] = const.CRM_CLUSTER_GEO_JOIN_H_OUTPUT
    cmd_help["crm_cluster_geo-init-arbitrator"] = const.CRM_CLUSTER_GEO_INIT_ARBIT_H_OUTPUT
    key = '_'.join(cmd.split())
    assert_eq(cmd_help[key], context.stdout)


@then('Corosync working on "{transport_type}" mode')
def step_impl(context, transport_type):
    if transport_type == "multicast":
        assert corosync.get_value("totem.transport") is None
    if transport_type == "unicast":
        assert_eq("udpu", corosync.get_value("totem.transport"))


@then('Expected votes will be "{votes}"')
def step_impl(context, votes):
    assert_eq(int(votes), int(corosync.get_value("quorum.expected_votes")))


@then('Directory "{directory}" created')
def step_impl(context, directory):
    assert os.path.isdir(directory) is True


@then('Directory "{directory}" not created')
def step_impl(context, directory):
    assert os.path.isdir(directory) is False


@then('Default crm_report tar file created')
def step_impl(context):
    default_file_name = 'crm_report-{}.tar.bz2'.format(datetime.datetime.now().strftime("%a-%d-%b-%Y"))
    assert os.path.exists(default_file_name) is True


@when('Remove default crm_report tar file')
def step_impl(context):
    default_file_name = 'crm_report-{}.tar.bz2'.format(datetime.datetime.now().strftime("%a-%d-%b-%Y"))
    os.remove(default_file_name)


@then('File "{f}" in "{archive}"')
def step_impl(context, f, archive):
    assert file_in_archive(f, archive) is True


@then('Directory "{f}" in "{archive}"')
def step_impl(context, f, archive):
    assert file_in_archive(f, archive) is True


@then('File "{f}" not in "{archive}"')
def step_impl(context, f, archive):
    assert file_in_archive(f, archive) is False


@then('File "{f}" was synced in cluster')
def step_impl(context, f):
    cmd = "crm cluster diff {}".format(f)
    rc, out, _ = run_command(context, cmd)
    assert_eq("", out)


@given('Resource "{res_id}" is started on "{node}"')
def step_impl(context, res_id, node):
    rc, out, err = ShellUtils().get_stdout_stderr("crm_mon -1")
    assert re.search(r'\*\s+{}\s+.*Started\s+{}'.format(res_id, node), out) is not None


@then('Resource "{res_id}" is started on "{node}"')
def step_impl(context, res_id, node):
    rc, out, err = ShellUtils().get_stdout_stderr("crm_mon -1")
    assert re.search(r'\*\s+{}\s+.*Started\s+{}'.format(res_id, node), out) is not None


@then('SBD option "{key}" value is "{value}"')
def step_impl(context, key, value):
    res = sbd.SBDUtils.get_sbd_value_from_config(key)
    assert_eq(value, res)


@then('SBD option "{key}" value for "{dev}" is "{value}"')
def step_impl(context, key, dev, value):
    res = sbd.SBDTimeout.get_sbd_msgwait(dev)
    assert_eq(int(value), res)


@then('Cluster property "{key}" is "{value}"')
def step_impl(context, key, value):
    res = crmutils.get_property(key)
    assert res is not None
    assert_eq(value.strip('s'), str(res).strip('s'))


@then('Property "{key}" in "{type}" is "{value}"')
def step_impl(context, key, type, value):
    res = crmutils.get_property(key, type)
    assert res is not None
    assert_eq(value.strip('s'), str(res).strip('s'))


@then('Parameter "{param_name}" not configured in "{res_id}"')
def step_impl(context, param_name, res_id):
    _, out, _ = run_command(context, "crm configure show {}".format(res_id))
    result = re.search("params .*{}=".format(param_name), out)
    assert result is None


@then('Parameter "{param_name}" configured in "{res_id}"')
def step_impl(context, param_name, res_id):
    _, out, _ = run_command(context, "crm configure show {}".format(res_id))
    result = re.search("params .*{}=".format(param_name), out)
    assert result is not None


@given('Yaml "{path}" value is "{value}"')
def step_impl(context, path, value):
    yaml_file = "/etc/crm/profiles.yml"
    with open(yaml_file) as f:
        data = yaml.load(f, Loader=yaml.SafeLoader)
    sec_name, key = path.split(':')
    assert_eq(str(value), str(data[sec_name][key]))


@when('Wait for DC')
def step_impl(context):
    while True:
        time.sleep(1)
        if crmutils.get_dc():
            break


@then('File "{path}" exists on "{node}"')
def step_impl(context, path, node):
    rc, _, stderr = behave_agent.call(node, 1122, 'test -f {}'.format(path), user='root')
    assert rc == 0


@given('File "{path}" not exist on "{node}"')
def step_impl(context, path, node):
    cmd = '[ ! -f {} ]'.format(path)
    rc, _, stderr = behave_agent.call(node, 1122, cmd, user='root')
    assert rc == 0


@then('File "{path}" not exist on "{node}"')
def step_impl(context, path, node):
    cmd = '[ ! -f {} ]'.format(path)
    rc, _, stderr = behave_agent.call(node, 1122, cmd, user='root')
    assert rc == 0


@then('Directory "{path}" is empty on "{node}"')
def step_impl(context, path, node):
    cmd = '[ ! "$(ls -A {})" ]'.format(path)
    rc, _, stderr = behave_agent.call(node, 1122, cmd, user='root')
    assert rc == 0


@then('Directory "{path}" not empty on "{node}"')
def step_impl(context, path, node):
    cmd = '[ "$(ls -A {})" ]'.format(path)
    rc, _, stderr = behave_agent.call(node, 1122, cmd, user='root')
    assert rc == 0


@then('Node "{node}" is UNCLEAN')
def step_impl(context, node):
    assert is_unclean(node) is True


@then('Wait "{count}" seconds for "{node}" successfully fenced')
def step_impl(context, count, node):
    index = 0
    while index <= int(count):
        rc, out, _ = ShellUtils().get_stdout_stderr("stonith_admin -h {}".format(node))
        if "Node {} last fenced at:".format(node) in out:
            return True
        time.sleep(1)
        index += 1
    return False

@then('Check passwordless for hacluster between "{nodelist}" "{result}"')
def step_impl(context, nodelist, result):
    assert result in ("successfully", "failed")
    if userdir.getuser() != 'root' or userdir.get_sudoer():
        return True
    failed = False
    nodes = nodelist.split()
    for i in range(0, len(nodes)):
        for j in range(i + 1, len(nodes)):
            rc, _, _ = behave_agent.call(
                nodes[i], 1122,
                f'ssh -o StrictHostKeyChecking=no -o ConnectTimeout=10 {nodes[j]} true',
                user='hacluster',
            )
            if rc != 0:
                failed = True
                context.logger.error(f"There is no passwordless configured from {nodes[i]} to {nodes[j]} under 'hacluster'")
    if result == "successfully":
        assert not failed
    else:
        assert failed


@then('Check user shell for hacluster between "{nodelist}"')
def step_impl(context, nodelist):
    if userdir.getuser() != 'root' or userdir.get_sudoer():
        return True
    for node in nodelist.split():
        if node == me():
            assert bootstrap.is_nologin('hacluster') is False
        else:
            assert bootstrap.is_nologin('hacluster', node) is False


@given('ssh-agent is started at "{path}" on nodes [{nodes:str+}]')
def step_impl(context, path, nodes):
    user =  userdir.get_sudoer()
    if not user:
        user = userdir.getuser()
    for node in nodes:
        rc, _, _ = behave_agent.call(node, 1122, f"systemd-run --uid '{user}' -u ssh-agent /usr/bin/ssh-agent -D -a '{path}'", user='root')
        assert 0 == rc


@then('This file "{target_file}" will trigger UnicodeDecodeError exception')
def step_impl(context, target_file):
    try:
        with open(target_file, "r", encoding="utf-8") as file:
            content = file.read()
    except UnicodeDecodeError as e:
        return True
    else:
        return False

@given('crm.conf poisoned on nodes [{nodes:str+}]')
def step_impl(context, nodes):
    for node in nodes:
        rc, _, _ = behave_agent.call(
            node, 1122,
            f'''mkdir -p /root/.config/crm && cat > /root/.config/crm/crm.conf << EOF
{const.CRM_CONF_CONTENT_POSIONED}
EOF''',
            user='root',
        )


@when('Set to previous schema version')
def step_impl(context):
    _, current_schema, _ = ShellUtils().get_stdout_stderr("crm configure schema")
    assert current_schema
    schema_list = ui_configure.schema_completer(None)
    assert schema_list
    assert current_schema in schema_list
    previous_schema = schema_list[schema_list.index(current_schema) - 1]
    context.previous_schema = previous_schema
    rc, _, _ = ShellUtils().get_stdout_stderr(f"crm configure schema {previous_schema}")
    assert rc == 0


@then('The schema version is the previous')
def step_impl(context):
    rc, schema, _ = ShellUtils().get_stdout_stderr("crm configure schema")
    assert rc == 0
    assert schema == context.previous_schema


@given('Get the latest schema version')
def step_impl(context):
    schema_list = ui_configure.schema_completer(None)
    assert schema_list
    context.schema_latest = schema_list[-1]


@when('Use crm configure upgrade to upgrade the schema')
def step_impl(context):
    rc, _, _ = ShellUtils().get_stdout_stderr("crm configure upgrade force")
    assert rc == 0


@then('The schema version is the latest')
def step_impl(context):
    rc, schema, _ = ShellUtils().get_stdout_stderr("crm configure schema")
    assert rc == 0
    assert schema == context.schema_latest


def firewall_cmd(context, node):
    rc, _, _ = run_command_local_or_remote(context, "firewall-cmd --state", node, exit_on_fail=False)
    return "firewall-cmd" if rc == 0 else "firewall-offline-cmd"


@then('Port "{port}" protocol "{proto}" is opened on "{node}"')
def step_impl(context, port, proto, node):
    cmd = f"{firewall_cmd(context, node)} --list-ports"
    rc, out, _ = run_command_local_or_remote(context, cmd, node, exit_on_fail=False)
    assert rc == 0
    assert f"{port}/{proto}" in out.split()


@given('The "{service}" firewalld service is available on "{node}"')
def step_impl(context, service, node):
    cmd = f"{firewall_cmd(context, node)} --info-service={service}"
    rc, _, _ = run_command_local_or_remote(context, cmd, node, exit_on_fail=False)
    assert rc == 0


def firewalld_list_services(context, node):
    cmd = f"{firewall_cmd(context, node)} --list-services"
    _, out, _ = run_command_local_or_remote(context, cmd, node, exit_on_fail=False)
    return out or ""


@then('The "{service}" firewalld service is added on "{node}"')
def step_impl(context, service, node):
    services = firewalld_list_services(context, node)
    assert service in services.split()


@then('The "{service}" firewalld service is not added on "{node}"')
def step_impl(context, service, node):
    services = firewalld_list_services(context, node)
    assert service not in services.split()


@given('The "{service}" firewalld service is not added on "{node}"')
def step_impl(context, service, node):
    services = firewalld_list_services(context, node)
    assert service not in services.split()
